# Copyright 2021-2022 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# flake8: noqa E402
import odb

import os
import sys
import locale
import inspect
import functools
from typing import Callable, Dict

# -- START: Environment Fixes
try:
    locale.setlocale(locale.LC_ALL, "C.UTF-8")
except locale.Error:
    # We tried. :)
    pass

if python_path := os.environ.get("ODB_PYTHONPATH", None):
    sys.path = python_path.split(":") + sys.path

sys.path.insert(0, os.path.dirname(os.environ["OPENLANE_ROOT"]))
# -- END

import click
import rich
from rich.table import Table
from openlane.state.design_format import DesignFormat, DesignFormatObject

click  # Re-export now that the environment actually works properly

write_fn: Dict[DesignFormat, Callable] = {
    DesignFormat.DEF: lambda reader, file: odb.write_def(reader.block, file),
    DesignFormat.ODB: lambda reader, file: odb.write_db(reader.db, file),
}


class OdbReader(object):
    def __init__(self, *args):
        self.db = odb.dbDatabase.create()
        if len(args) == 1:
            db_in = args[0]
            self.db = odb.read_db(self.db, db_in)
        elif len(args) == 2:
            lef_in, def_in = args
            if not (isinstance(lef_in, list) or isinstance(lef_in, tuple)):
                lef_in = [lef_in]
            for lef in lef_in:
                odb.read_lef(self.db, lef)
            if def_in is not None:
                odb.read_def(self.db, def_in)

        self.tech = self.db.getTech()
        self.chip = self.db.getChip()
        if self.chip is not None:
            self.block = self.db.getChip().getBlock()
            self.name = self.block.getName()
            self.rows = self.block.getRows()
            self.dbunits = self.block.getDefUnits()
            self.instances = self.block.getInsts()

    def add_lef(self, new_lef):
        odb.read_lef(self.db, new_lef)


def click_odb(function):
    @functools.wraps(function)
    def wrapper(input_db, input_lefs, **kwargs):
        reader = OdbReader(input_db)

        signature = inspect.signature(function)
        parameter_keys = signature.parameters.keys()

        kwargs = kwargs.copy()
        kwargs["reader"] = reader

        outputs = []
        for key, value in kwargs.items():
            if key.startswith("output_"):
                id = key[7:]
                outputs.append((DesignFormat.by_id(id), value))

        kwargs = {k: kwargs[k] for k in kwargs.keys() if not k.startswith("output_")}

        if "input_db" in parameter_keys:
            kwargs["input_db"] = input_db
        if "input_lefs" in parameter_keys:
            kwargs["input_lefs"] = input_lefs

        if input_db.endswith(".def"):
            print(
                "Error: Invocation was not updated to use an odb file.", file=sys.stderr
            )
            exit(1)

        function(**kwargs)

        for format, path in outputs:
            fn = write_fn[format]
            fn(reader, path)

    for format in DesignFormat:
        if write_fn.get(format) is None:
            continue
        # For type-checker: all guaranteed to be DesignFormatObjects
        assert isinstance(format.value, DesignFormatObject)
        wrapper = click.option(
            f"--output-{format.value.id}",
            default=None,
            help=f"Write {format.value.name}",
        )(wrapper)

    wrapper = click.option(
        "-l",
        "--input-lef",
        "input_lefs",
        default=(),
        help="LEF file needed to have a proper view of the DEF files",
        multiple=True,
    )(wrapper)
    wrapper = click.argument("input_db")(wrapper)

    return wrapper



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from enum import Enum
from dataclasses import dataclass
from typing import Dict, Optional


@dataclass
class DesignFormatObject:
    id: str
    extension: str
    name: str
    folder_override: Optional[str] = None
    multiple: bool = False

    @property
    def folder(self) -> str:
        return self.folder_override or self.id


class DesignFormat(Enum):
    """
    A list of design formats that may be kept in the state object.
    """

    NETLIST: DesignFormatObject = DesignFormatObject(
        "nl",
        "nl.v",
        "Verilog Netlist",
    )
    POWERED_NETLIST: DesignFormatObject = DesignFormatObject(
        "pnl",
        "pnl.v",
        "Powered Verilog Netlist",
    )

    DEF: DesignFormatObject = DesignFormatObject(
        "def",
        "def",
        "Design Exchange Format",
    )
    LEF: DesignFormatObject = DesignFormatObject(
        "lef",
        "lef",
        "Library Exchange Format",
    )
    ODB: DesignFormatObject = DesignFormatObject(
        "odb",
        "odb",
        "OpenDB Database",
    )

    SDC: DesignFormatObject = DesignFormatObject(
        "sdc",
        "sdc",
        "Design Constraints",
    )
    SDF: DesignFormatObject = DesignFormatObject(
        "sdf",
        "sdf",
        "Standard Delay Format",
    )
    SPEF: DesignFormatObject = DesignFormatObject(
        "spef",
        "spef",
        "Standard Parasitics Extraction Format",
        multiple=True,  # nom, min, max, ...
    )
    LIB: DesignFormatObject = DesignFormatObject(
        "lib",
        "lib",
        "LIB Timing Library Format",
        multiple=True,
    )
    SPICE: DesignFormatObject = DesignFormatObject(
        "spice",
        "spice",
        "Simulation Program with Integrated Circuit Emphasis",
    )

    GDS: DesignFormatObject = DesignFormatObject(
        "gds",
        "gds",
        "GDSII Stream",
    )
    MAG_GDS: DesignFormatObject = DesignFormatObject(
        "mag_gds",
        "magic.gds",
        "GDSII Stream (Magic)",
    )
    KLAYOUT_GDS: DesignFormatObject = DesignFormatObject(
        "klayout_gds",
        "klayout.gds",
        "GDSII Stream (KLayout)",
    )

    JSON_HEADER: DesignFormatObject = DesignFormatObject(
        "json_h",
        "h.json",
        "Design JSON Header File",
    )

    @staticmethod
    def by_id(id: str) -> Optional["DesignFormat"]:
        return _designformat_by_id.get(id)


_designformat_by_id: Dict[str, "DesignFormat"] = {
    format.value.id: format for format in DesignFormat
}



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from functools import reduce
import os
import re
import shutil
import sys
import json
from decimal import Decimal
from abc import abstractmethod
from typing import List, Optional

from .step import Step, StepException
from .common_variables import io_layer_variables
from ..logging import warn
from ..state import State, DesignFormat, Path
from ..config import Variable, StringEnum, Macro
from ..common import get_openlane_root, get_script_dir

inf_rx = re.compile(r"\b(-?)inf\b")


class OdbpyStep(Step):
    inputs = [DesignFormat.ODB]
    outputs = [DesignFormat.ODB, DesignFormat.DEF]

    def run(self, state_in, **kwargs) -> State:
        state_out = super().run(state_in, **kwargs)
        kwargs, env = self.extract_env(kwargs)

        out_paths = {}

        command = self.get_command()
        for output in [DesignFormat.ODB, DesignFormat.DEF]:
            filename = f"{self.config['DESIGN_NAME']}.{output.value.extension}"
            file_path = os.path.join(self.step_dir, filename)
            command.append(f"--output-{output.value.id}")
            command.append(file_path)
            out_paths[output] = Path(file_path)

        command += [
            str(state_out[DesignFormat.ODB]),
        ]

        env["OPENLANE_ROOT"] = get_openlane_root()
        env["ODB_PYTHONPATH"] = ":".join(sys.path)

        self.run_subprocess(
            command,
            env=env,
            **kwargs,
        )

        metrics_path = os.path.join(self.step_dir, "or_metrics_out.json")
        if os.path.exists(metrics_path):
            metrics_str = open(metrics_path).read()
            metrics_str = inf_rx.sub(lambda m: f"{m[1] or ''}\"Infinity\"", metrics_str)
            new_metrics = json.loads(metrics_str)
            state_out.metrics.update(new_metrics)

        for output in [DesignFormat.ODB, DesignFormat.DEF]:
            state_out[output] = out_paths[output]

        return state_out

    def get_command(self) -> List[str]:
        metrics_path = os.path.join(self.step_dir, "or_metrics_out.json")

        tech_lefs = self.toolbox.filter_views(self.config, self.config["TECH_LEFS"])
        if len(tech_lefs) != 1:
            raise StepException(
                "Misconfigured SCL: 'TECH_LEFS' must return exactly one Tech LEF for its default timing corner."
            )

        lefs = ["--input-lef", tech_lefs[0]]
        for lef in self.config["CELL_LEFS"]:
            lefs.append("--input-lef")
            lefs.append(lef)
        if extra_lefs := self.config["EXTRA_LEFS"]:
            for lef in extra_lefs:
                lefs.append("--input-lef")
                lefs.append(lef)
        return (
            [
                "openroad",
                "-exit",
                "-metrics",
                metrics_path,
                "-python",
                self.get_script_path(),
            ]
            + self.get_subcommand()
            + lefs
        )

    @abstractmethod
    def get_script_path(self):
        pass

    def get_subcommand(self) -> List[str]:
        return []


@Step.factory.register()
class ApplyDEFTemplate(OdbpyStep):
    """
    Copies the floorplan of a "template" DEF file for a new design, i.e.,
    it will copy the die area, core area, and non-power pin names and locations.
    """

    id = "Odb.ApplyDEFTemplate"
    name = "Apply DEF Template"

    flow_control_variable = "FP_DEF_TEMPLATE"
    flow_control_msg = "No DEF template provided, skipping…"

    config_vars = [
        Variable(
            "FP_DEF_TEMPLATE",
            Optional[Path],
            "Points to the DEF file to be used as a template.",
        ),
    ]

    def get_script_path(self):
        return os.path.join(
            get_script_dir(),
            "odbpy",
            "apply_def_template.py",
        )

    def get_command(self) -> List[str]:
        return super().get_command() + [
            "--def-template",
            self.config["FP_DEF_TEMPLATE"],
        ]


class SetPowerConnections(OdbpyStep):
    id = "Odb.SetPowerConnections"
    name = "Set Power Connections"
    inputs = [DesignFormat.JSON_HEADER, DesignFormat.ODB]

    def get_script_path(self):
        return os.path.join(get_script_dir(), "odbpy", "set_power_connections.py")

    def get_command(self) -> List[str]:
        state_in = self.state_in.result()
        return super().get_command() + [
            "--input-json",
            str(state_in[DesignFormat.JSON_HEADER]),
        ]


@Step.factory.register()
class ManualMacroPlacement(OdbpyStep):
    """
    Performs macro placement using a simple configuration file. The file is
    defined as a line-break delimited list of instances and positions, in the
    format ``instance_name X_pos Y_pos Orientation``.

    If no macro instances are configured, this step is skipped.
    """

    id = "Odb.ManualMacroPlacement"
    name = "Manual Macro Placement"

    config_vars = [
        Variable(
            "MACRO_PLACEMENT_CFG",
            Optional[Path],
            "Path to an optional override for instance placement instead of the `MACROS` object for compatibility with OpenLane 1. If both are `None`, this step is skipped.",
        ),
    ]

    def get_script_path(self):
        return os.path.join(get_script_dir(), "odbpy", "manual_macro_place.py")

    def get_command(self) -> List[str]:
        return super().get_command() + [
            "--config",
            os.path.join(self.step_dir, "placement.cfg"),
            "--fixed",
        ]

    def run(self, state_in: State, **kwargs) -> State:
        cfg_file = Path(os.path.join(self.step_dir, "placement.cfg"))
        if cfg_ref := self.config.get("MACRO_PLACEMENT_CFG"):
            warn(
                "Using 'MACRO_PLACEMENT_CFG' is deprecated. It is recommended to use the 'MACROS' object."
            )
            shutil.copyfile(cfg_ref, cfg_file)
        elif macros := self.config.get("MACROS"):
            instance_count = reduce(
                lambda x, y: x + len(y.instances), macros.values(), 0
            )
            if instance_count >= 1:
                with open(cfg_file, "w") as f:
                    for module, macro in macros.items():
                        if not isinstance(macro, Macro):
                            raise StepException(
                                f"Misconstructed configuration: macro definition for key {module} is not of type 'Macro'."
                            )
                        for name, data in macro.instances.items():
                            f.write(
                                f"{name} {data.location[0]} {data.location[1]} {data.orientation}\n"
                            )

        if not cfg_file.exists():
            warn("No instances found, skipping…")
            return Step.run(self, state_in, **kwargs)

        return super().run(state_in, **kwargs)


@Step.factory.register()
class ReportWireLength(OdbpyStep):
    """
    Outputs a CSV of long wires, printed by length. Useful as a design aid to
    detect when one wire is connected to too many things.
    """

    outputs = []

    id = "Odb.ReportWireLength"
    name = "Report Wire Length"
    outputs = []

    def get_script_path(self):
        return os.path.join(get_script_dir(), "odbpy", "wire_lengths.py")

    def get_command(self) -> List[str]:
        return super().get_command() + [
            "--human-readable",
            "--report-out",
            os.path.join(self.step_dir, "wire_lengths.csv"),
        ]


@Step.factory.register()
class ReportDisconnectedPins(OdbpyStep):
    id = "Odb.ReportDisconnectedPins"
    name = "Report Disconnected Pins"

    def get_script_path(self):
        return os.path.join(get_script_dir(), "odbpy", "disconnected_pins.py")

    def get_command(self) -> List[str]:
        command = super().get_command()
        if ignored_modules := self.config["IGNORE_DISCONNECTED_MODULES"]:
            for module in ignored_modules:
                command.append("--ignore-module")
                command.append(module)
        return command


@Step.factory.register()
class CustomIOPlacement(OdbpyStep):
    """
    Places I/O pins using a custom script, which uses a "pin order configuration"
    file.

    Check the reference documentation for the structure of said file.
    """

    id = "Odb.CustomIOPlacement"
    name = "Custom I/O Placement"
    long_name = "Custom I/O Pin Placement Script"

    flow_control_variable = "FP_PIN_ORDER_CFG"
    flow_control_msg = "No custom floorplan file configured, skipping…"

    config_vars = io_layer_variables + [
        Variable(
            "FP_PIN_ORDER_CFG",
            Optional[Path],
            "Path to the configuration file. If set to `None`, this step is skipped.",
        ),
        Variable(
            "QUIT_ON_UNMATCHED_IO",
            bool,
            "Exit on unmatched pins in a provided `FP_PIN_ORDER_CFG` file.",
            default=True,
            deprecated_names=["FP_IO_UNMATCHED_ERROR"],
        ),
    ]

    def get_script_path(self):
        return os.path.join(get_script_dir(), "odbpy", "io_place.py")

    def get_command(self) -> List[str]:
        length = max(
            self.config["FP_IO_VLENGTH"],
            self.config["FP_IO_HLENGTH"],
        )

        return super().get_command() + [
            "--config",
            self.config["FP_PIN_ORDER_CFG"],
            "--hor-layer",
            self.config["FP_IO_HLAYER"],
            "--ver-layer",
            self.config["FP_IO_VLAYER"],
            "--hor-width-mult",
            str(self.config["FP_IO_VTHICKNESS_MULT"]),
            "--ver-width-mult",
            str(self.config["FP_IO_HTHICKNESS_MULT"]),
            "--hor-extension",
            str(self.config["FP_IO_HEXTEND"]),
            "--ver-extension",
            str(self.config["FP_IO_VEXTEND"]),
            "--length",
            str(length),
            (
                "--unmatched-error"
                if self.config["QUIT_ON_UNMATCHED_IO"]
                else "--ignore-unmatched"
            ),
        ]


@Step.factory.register()
class DiodesOnPorts(OdbpyStep):
    """
    Unconditionally inserts diodes on design ports diodes on ports,
    to mitigate the `antenna effect <https://en.wikipedia.org/wiki/Antenna_effect>`_.

    Useful for hardening macros, where ports may get long wires that are
    unaccounted for when hardening a top-level chip.
    """

    id = "Odb.DiodesOnPorts"
    name = "Diodes on Ports"
    long_name = "Diode on Port Insertion Script"

    config_vars = [
        Variable(
            "DIODE_ON_PORTS",
            StringEnum("DIODE_ON_PORTS", ["none", "in", "out", "both"]),
            "Always insert diodes on ports with the specified polarities.",
            default="none",
        ),
    ]

    def get_script_path(self):
        return os.path.join(get_script_dir(), "odbpy", "diodes.py")

    def get_subcommand(self) -> List[str]:
        return ["place"]

    def get_command(self) -> List[str]:
        cell, pin = self.config["DIODE_CELL"].split("/")

        return super().get_command() + [
            "--threshold",
            "Infinity",
            "--diode-cell",
            cell,
            "--diode-pin",
            pin,
            "--port-protect",
            self.config["DIODE_ON_PORTS"].value,
        ]

    def run(self, state_in: State, **kwargs) -> State:
        if self.config["DIODE_ON_PORTS"].value == "none":
            warn("'DIODE_ON_PORTS' is set to 'none': skipping…")
            return Step.run(self, state_in, **kwargs)

        if self.config["GPL_CELL_PADDING"] == 0:
            warn(
                "'GPL_CELL_PADDING' is set to 0. This step may cause overlap failures."
            )

        return super().run(state_in, **kwargs)


@Step.factory.register()
class HeuristicDiodeInsertion(OdbpyStep):
    """
    Runs a custom diode insertion script to mitigate the `antenna effect <https://en.wikipedia.org/wiki/Antenna_effect>`_.

    This script uses the `Manhattan length <https://en.wikipedia.org/wiki/Manhattan_distance>`_
    of a (non-existent) wire at the global placement stage, and places diodes
    if they exceed a certain threshold. This, however, requires some padding:
    `GPL_CELL_PADDING` and `DPL_CELL_PADDING` must be higher than 0 for this
    script to work reliably.

    This step is unique in that it is the only step with a ``RUN_`` variable that
    is disabled by default. This is for compatibility with OpenLane 1 configs.

    The original script was written by `Sylvain "tnt" Munaut <https://github.com/smunaut>`_.
    """

    id = "Odb.HeuristicDiodeInsertion"
    name = "Heuristic Diode Insertion"
    long_name = "Heuristic Diode Insertion Script"

    flow_control_variable = "RUN_HEURISTIC_DIODE_INSERTION"
    config_vars = [
        Variable(
            "RUN_HEURISTIC_DIODE_INSERTION",
            bool,
            "Enables/disables this step.",
            default=False,  # For compatibility with OL1. Yep.
        ),
        Variable(
            "HEURISTIC_ANTENNA_THRESHOLD",
            Optional[Decimal],
            "A manhattan distance above which a diode is recommended to be inserted by a heuristic inserter. If not specified, the heuristic inserter will typically use a default value.",
            units="µm",
        ),
    ]

    def get_script_path(self):
        return os.path.join(get_script_dir(), "odbpy", "diodes.py")

    def get_subcommand(self) -> List[str]:
        return ["place"]

    def get_command(self) -> List[str]:
        cell, pin = self.config["DIODE_CELL"].split("/")

        threshold_opts = []
        if threshold := self.config["HEURISTIC_ANTENNA_THRESHOLD"]:
            threshold_opts = ["--threshold", threshold]

        return (
            super().get_command()
            + [
                "--diode-cell",
                cell,
                "--diode-pin",
                pin,
            ]
            + threshold_opts
        )

    def run(self, state_in: State, **kwargs) -> State:
        if self.config["GPL_CELL_PADDING"] == 0:
            warn(
                "'GPL_CELL_PADDING' is set to 0. This step may cause overlap failures."
            )

        return super().run(state_in, **kwargs)



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import tkinter
import tempfile
from decimal import Decimal, InvalidOperation

from .config import Config


def env_from_tcl(env_in: Config, tcl_in: str) -> Config:
    interpreter = tkinter.Tcl()

    initial_env = os.environ.copy()
    env_out = env_in.copy()._unlock()

    with tempfile.NamedTemporaryFile("r+") as f:
        env_str = ""
        unset_env_str = ""
        for key, value in env_in.items():
            env_str += f"set ::env({key}) {{{value}}}\n"
            unset_env_str += f"unset ::env({key})\n"

        tcl_script = f"""
        {env_str}
        {tcl_in}
        set f [open {f.name} WRONLY]
        foreach key [array names ::env] {{
            puts $f "$key $::env($key)\\0"
        }}
        close $f
        {unset_env_str}
        """
        interpreter.eval(tcl_script)

        f.seek(0)
        env_strings = f.read()

        for line in env_strings.split("\0\n"):
            if line.strip() == "":
                continue
            key, value = line.split(" ", 1)
            try:
                value = Decimal(value)
            except InvalidOperation:
                pass
            if initial_env.get(key) is None and env_in.get(key) != value:
                env_out[key] = value

    return env_out._lock()



# Copyright 2020-2022 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import odb

from reader import OdbReader, click_odb, click


@click.command()
@click.option(
    "-D",
    "--top-def",
    required=True,
    help="DEF view of the top-level design instantiating the macro",
)
@click.option(
    "-L",
    "--top-lef",
    required=True,
    help="DEF view of the top-level design instantiating the macro",
)
@click.option("--keep-inner-connections", default=False, is_flag=True)
@click_odb
def contextualize(reader, top_def, top_lef, keep_inner_connections):
    reader.add_lef(top_lef)

    top = OdbReader(top_lef, top_def)
    macro = reader

    print("Block design name:", macro.name)
    print("Top-level design name:", top.name)

    nets_top = top.block.getNets()
    to_connect = {}
    MACRO_TOP_PLACEMENT_X = 0
    MACRO_TOP_PLACEMENT_Y = 0

    assert macro.name in [
        inst.getMaster().getName() for inst in top.block.getInsts()
    ], "%s not found in %s" % (macro.name, top.name)

    for net in nets_top:
        iterms = net.getITerms()  # asssumption: no pins (bterms) on top level
        block_net_name = None
        for iterm in iterms:
            macro_name = iterm.getMTerm().getMaster().getName()
            if macro_name == macro.name:
                block_net_name = iterm.getMTerm().getName()
                macro_top_inst = iterm.getInst()
                (
                    MACRO_TOP_PLACEMENT_X,
                    MACRO_TOP_PLACEMENT_Y,
                ) = macro_top_inst.getLocation()
                print("Block net name: ", block_net_name)
                break
        if block_net_name is not None:
            to_connect[block_net_name] = []
            for iterm in iterms:
                macro_name = iterm.getMTerm().getMaster().getName()
                if macro_name != macro.name:
                    to_connect[block_net_name].append(iterm)
            block_net_name = None

            # print(macro_name, inst_name, end= ' ')
            # print(iterm.getMTerm().getName())

    # print(to_connect)

    nets_macro = macro.block.getNets()
    created_macros = {}
    for net in nets_macro:
        iterms = net.getITerms()  # asssumption: no pins (bterms) on top level
        if not keep_inner_connections:
            for iterm in iterms:
                iterm.disconnect()
        if net.getName() in to_connect:
            for node_iterm in to_connect[net.getName()]:
                node_master = node_iterm.getMTerm().getMaster()
                node_inst = node_iterm.getInst()
                node_inst_name = node_iterm.getInst().getName()
                if node_inst_name not in created_macros:
                    created_macros[node_inst_name] = 1
                    print("Creating: ", node_master.getName(), node_inst_name)
                    new_inst = odb.dbInst_create(
                        macro.block, node_master, node_inst_name
                    )
                    new_inst.setOrient(node_inst.getOrient())
                    new_inst.setLocation(
                        node_inst.getLocation()[0] - MACRO_TOP_PLACEMENT_X,
                        node_inst.getLocation()[1] - MACRO_TOP_PLACEMENT_Y,
                    )
                    new_inst.setPlacementStatus("FIRM")
                else:
                    new_inst = macro.block.findInst(node_inst_name)
                new_inst.findITerm(node_iterm.getMTerm().getName()).connect(net)


if __name__ == "__main__":
    contextualize()



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import sys
import glob
import shutil
import marshal
import tempfile
import subprocess
from textwrap import dedent
from functools import partial
import traceback
from typing import Tuple, Type, Optional, List, Union

import click
from cloup import (
    option,
    option_group,
    command,
    version_option,
    HelpFormatter,
    HelpTheme,
    Style,
)
from cloup.constraints import (
    If,
    IsSet,
    accept_none,
    require_one,
    mutually_exclusive,
)

from .__version__ import __version__
from .state import State
from .logging import (
    LogLevelsDict,
    set_log_level,
    err,
    warn,
    info,
)
from .common import (
    get_opdks_rev,
    get_openlane_root,
)
from .container import run_in_container, sanitize_path
from .flows import Flow, SequentialFlow, FlowException, FlowError
from .config import Config, InvalidConfig
from .plugins import discovered_plugins


def run(
    ctx: click.Context,
    flow_name: Optional[str],
    use_volare: bool,
    pdk_root: Optional[str],
    pdk: str,
    scl: Optional[str],
    config_files: str,
    run_tag: Optional[str],
    last_run: bool,
    frm: Optional[str],
    to: Optional[str],
    only: Optional[str],
    skip: Tuple[str, ...],
    initial_state_json: Optional[str],
    config_override_strings: List[str],
    log_level: str,
):
    try:
        set_log_level(log_level)
    except ValueError:
        err(f"Invalid logging level: {log_level}.")
        click.echo(ctx.get_help())
        exit(-1)

    if only is not None:
        frm = to = only

    if use_volare:
        import volare

        pdk_root = volare.get_volare_home(pdk_root)

        volare.enable(pdk_root, pdk[:-1], get_opdks_rev())

    config_file = config_files[0]

    # Enforce Mutual Exclusion
    if run_tag is not None and last_run:
        err("--run-tag and --last-run are mutually exclusive.")
        exit(-1)

    flow_description: Union[str, List[str]] = flow_name or "Classic"

    if meta := Config.get_meta(config_file, flow_override=flow_name):
        if flow_ids := meta.flow:
            flow_description = flow_ids

    TargetFlow: Type[Flow]

    if not isinstance(flow_description, str):
        TargetFlow = SequentialFlow.make(flow_description)
    else:
        if FlowClass := Flow.get(flow_description):
            TargetFlow = FlowClass
        else:
            err(
                f"Unknown flow '{flow_description}' specified in configuration file's 'meta' object."
            )
            exit(-1)

    try:
        flow = TargetFlow.init_with_config(
            config_in=config_file,
            pdk_root=pdk_root,
            pdk=pdk,
            scl=scl,
            config_override_strings=config_override_strings,
        )
    except InvalidConfig as e:
        info(f"[green]Errors have occurred while loading the {e.config}.")
        for error in e.errors:
            err(error)

        if len(e.warnings) > 0:
            info("The following warnings have also been generated:")
            for warning in e.warnings:
                warn(warning)
        info("OpenLane will now quit. Please check your configuration.")
        exit(1)
    except ValueError as e:
        err(e)
        info("OpenLane will now quit.")
        exit(1)

    initial_state: Optional[State] = None
    if initial_state_json is not None:
        initial_state = State.loads(open(initial_state_json).read())

    if last_run:
        runs = glob.glob(os.path.join(flow.design_dir, "runs", "*"))

        latest_time: float = 0
        latest_run: Optional[str] = None
        for run in runs:
            time = os.path.getmtime(run)
            if time > latest_time:
                latest_time = time
                latest_run = run

        if latest_run is None:
            err("--last-run specified, but no runs found.")
            exit(1)

        run_tag = os.path.basename(latest_run)

    try:
        flow.start(
            tag=run_tag,
            frm=frm,
            to=to,
            skip=list(skip),
            with_initial_state=initial_state,
        )
    except FlowException as e:
        err(f"The flow has encountered an unexpected error: {e}")
        err("OpenLane will now quit.")
        exit(1)
    except FlowError as e:
        if "deferred" not in str(e):
            err(f"The following error was encountered while running the flow: {e}")
        err("OpenLane will now quit.")
        exit(2)


def print_bare_version(
    ctx: click.Context,
    param: click.Parameter,
    value: bool,
):
    if not value:
        return
    print(__version__, end="")
    ctx.exit(0)


def run_smoke_test(
    ctx: click.Context,
    param: click.Parameter,
    value: bool,
):
    if not value:
        return
    dockerized = ctx.params["dockerized"]

    with tempfile.TemporaryDirectory("_ol2design") as d:
        final_path = os.path.join(d, "spm")
        shutil.copytree(
            os.path.join(get_openlane_root(), "smoke_test_design"),
            final_path,
        )

        cmd = [
            (sys.executable if not dockerized else "python3"),
            "-m",
            "openlane",
            os.path.join(final_path, "config.json"),
        ]
        if pdk_root := ctx.params.get("pdk_root"):
            cmd += ["--pdk-root", pdk_root]

        if dockerized:
            cmd += ["--dockerized", "--docker-mount", d]
            if extra_mounts := ctx.params.get("docker_mounts"):
                for mount in extra_mounts:
                    cmd += ["--docker-mount", mount]

        try:
            subprocess.check_call(cmd)
        except subprocess.CalledProcessError:
            print("")
            err("Smoke test failed.")
            ctx.exit(-1)

    ctx.exit(0)


def cli_in_container(ctx: click.Context, docker_mounts: Tuple[str], kwargs: dict):
    f = None
    status = 0
    try:
        if isj := kwargs.get("initial_state_json"):
            _, kwargs["initial_state_json"] = sanitize_path(isj)

        config_files: List[str] = []
        for file in kwargs["config_files"]:
            _, target = sanitize_path(file)
            config_files.append(target)
        kwargs["config_files"] = config_files

        pdk_root = kwargs.pop("pdk_root")
        pdk = kwargs.pop("pdk")
        scl = kwargs.pop("scl")

        f = tempfile.NamedTemporaryFile("wb", suffix=".marshalled", delete=False)
        marshal.dump(kwargs, f, 4)
        f.close()
        _, binary = sanitize_path(f.name)

        run_in_container(
            f"ghcr.io/efabless/openlane2:{__version__}",
            ["python3", "-m", "openlane", binary],
            pdk_root=pdk_root,
            pdk=pdk,
            scl=scl,
            other_mounts=docker_mounts + (os.path.dirname(f.name),),
        )
    except ValueError as e:
        print(e)
        status = -1
    except subprocess.CalledProcessError as e:
        status = e.returncode
    except Exception:
        traceback.print_exc()
        status = -1
    finally:
        if f is not None:
            try:
                os.unlink(f.name)
            except FileNotFoundError:
                pass
        ctx.exit(status)


def print_plugins(
    ctx: click.Context,
    param: click.Parameter,
    value: bool,
):
    if not value:
        return
    print(f"openlane -> {__version__}")
    for name, module in discovered_plugins.items():
        print(f"{name} -> {module.__version__}")
    ctx.exit(0)


formatter_settings = HelpFormatter.settings(
    theme=HelpTheme(
        invoked_command=Style(fg="bright_yellow"),
        heading=Style(fg="bright_white", bold=True),
        constraint=Style(fg="magenta"),
        col1=Style(fg="bright_yellow"),
    )
)

o = partial(option, show_default=True)


@command(
    no_args_is_help=True,
    formatter_settings=formatter_settings,
)
@version_option(
    __version__,
    prog_name="OpenLane",
    message=dedent(
        """
        %(prog)s v%(version)s

        Copyright ©2020-2023 Efabless Corporation and other contributors.

        Available under the Apache License, version 2. Included with the source code,
        but you can also get a copy at https://www.apache.org/licenses/LICENSE-2.0

        Included tools and utilities may be distributed under stricter licenses.
        """
    ).strip(),
)
@o(
    "--bare-version",
    is_flag=True,
    is_eager=True,
    expose_value=False,
    callback=print_bare_version,
    hidden=True,
)
@o(
    "--list-plugins",
    is_flag=True,
    is_eager=True,
    expose_value=False,
    callback=print_plugins,
    help="List all detected plugins for OpenLane.",
)
@option_group(
    "PDK options",
    o(
        "--volare-pdk/--manual-pdk",
        "use_volare",
        default=True,
        help="Automatically use Volare for PDK version installation and enablement. Set --manual if you want to use a custom PDK version.",
    ),
    o(
        "--pdk-root",
        is_eager=True,
        default=os.environ.pop("PDK_ROOT", None),
        help="Override volare PDK root folder. Required if Volare is not installed.",
    ),
    o(
        "-p",
        "--pdk",
        type=str,
        default=os.environ.pop("PDK", "sky130A"),
        help="The process design kit to use.",
    ),
    o(
        "-s",
        "--scl",
        type=str,
        default=os.environ.pop("STD_CELL_LIBRARY", None),
        help="The standard cell library to use. If None, the PDK's default standard cell library is used.",
    ),
)
@option_group(
    "Docker options",
    o(
        "--dockerized/--native",
        is_eager=True,
        default=False,
        help="Run OpenLane primarily using a Docker container. Some caveats apply.",
    ),
    o(
        "--docker-mount",
        "docker_mounts",
        multiple=True,
        is_eager=True,
        default=[],
        help="Mount this directory in dockerized mode. Can be supplied multiple times to mount multiple directories.",
    ),
    constraint=If(~IsSet("dockerized"), accept_none),
)
@option_group(
    "Run options",
    o(
        "-f",
        "--flow",
        "flow_name",
        type=click.Choice(Flow.factory.list(), case_sensitive=False),
        default=None,
        help="The built-in OpenLane flow to use for this run",
    ),
    o(
        "-i",
        "--with-initial-state",
        "initial_state_json",
        type=click.Path(
            exists=True,
            file_okay=True,
            dir_okay=False,
        ),
        default=None,
        help="Use this JSON file as an initial state. If this is not specified, the latest `state_out.json` of the run directory will be used if available.",
    ),
    o(
        "-c",
        "--override-config",
        "config_override_strings",
        type=str,
        multiple=True,
        help="For this run only- override a configuration variable with a certain value. In the format KEY=VALUE. Can be specified multiple times. Values must be valid JSON values.",
    ),
)
@option_group(
    "Run options - tag",
    o(
        "--run-tag",
        default=None,
        type=str,
        help="An optional name to use for this particular run of an OpenLane-based flow.",
    ),
    o(
        "--last-run",
        is_flag=True,
        default=False,
        help="Use the last run as the run tag.",
    ),
    constraint=mutually_exclusive,
)
@o(
    "--smoke-test",
    is_flag=True,
    help="Runs a basic OpenLane smoke test.",
    expose_value=False,
    callback=run_smoke_test,
)
@option_group(
    "Run options - sequential flow control",
    o(
        "-F",
        "--from",
        "frm",
        type=str,
        default=None,
        help="Start from a step with this id. Supported by sequential flows.",
    ),
    o(
        "-T",
        "--to",
        type=str,
        default=None,
        help="Stop at a step with this id. Supported by sequential flows.",
    ),
    o(
        "--only",
        type=str,
        default=None,
        help="Shorthand to set both --from and --to to the same value.",
    ),
    o(
        "-S",
        "--skip",
        type=str,
        multiple=True,
        help="Skip these steps. Supported by sequential flows.",
    ),
    constraint=If(IsSet("only"), require_one),
)
@o(
    "--log-level",
    type=str,
    default="VERBOSE",
    help=dedent(
        """
        A logging level. Set to INFO or higher to silence subprocess logs.

        You can provide either a number or a string out of the following (higher is more silent):
        """
    )
    + ",".join([f"{name}={value}" for name, value in LogLevelsDict.items()]),
)
@click.argument(
    "config_files",
    required=True,
    nargs=-1,
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=True,
    ),
)
@click.pass_context
def cli(ctx: click.Context, dockerized: bool, docker_mounts: Tuple[str], **kwargs):
    if dockerized:
        cli_in_container(ctx, docker_mounts, kwargs)
    run_kwargs = kwargs
    args = kwargs["config_files"]
    if len(args) == 1 and args[0].endswith(".marshalled"):
        run_kwargs = marshal.load(open(args[0], "rb"))
        run_kwargs.update(**{k: kwargs[k] for k in ["pdk_root", "pdk", "scl"]})
    run(ctx, **run_kwargs)


if __name__ == "__main__":
    cli()



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The State Module

This module manages the State of a Design before and after the execution of an
OpenLane step. The State is essentially a list of views in various formats in
addition to the cumulative set of metrics created by a given step.
"""
from .design_format import DesignFormat, DesignFormatObject
from .state import Path, State, InvalidState



# Copyright 2022 Efabless Corporation
#
# This file is part of the DFFRAM Memory Compiler.
# See https://github.com/Cloud-V/DFFRAM for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

import odb

from reader import click, click_odb, OdbReader
from reader import rich
from reader import Table

import utl


def is_connected(iterm: odb.dbITerm) -> bool:
    return iterm.getNet() is not None


def is_special(iterm: odb.dbITerm) -> bool:
    return iterm.getSigType() == "POWER" or iterm.getSigType() == "GROUND"


def has_disconnect(instance: odb.dbInst) -> bool:
    return (
        True
        if [iterm for iterm in instance.getITerms() if not is_connected(iterm)]
        else False
    )


@click.option(
    "--ignore-module", default=[""], multiple=True, type=str, help="Modules to ignore"
)
@click.command()
@click_odb
def main(
    reader: OdbReader,
    ignore_module: List[str],
):
    db = reader.db
    block = db.getChip().getBlock()
    instances = block.getInsts()
    table = Table(
        "Instance", "Power", "Disconnected", "Signal", "Disconnected", title=""
    )
    disconnected_pins_count = 0

    instances_with_disconnect = [
        instance
        for instance in instances
        if has_disconnect(instance)
        and (instance.getMaster().getName() not in ignore_module)
    ]

    for instance in instances_with_disconnect:
        iterms = instance.getITerms()
        signal_pins = [
            iterm.getMTerm().getName() for iterm in iterms if not is_special(iterm)
        ]
        power_pins = [
            iterm.getMTerm().getName() for iterm in iterms if is_special(iterm)
        ]
        disconnected_power_pins = [
            iterm.getMTerm().getName()
            for iterm in iterms
            if is_special(iterm) and not is_connected(iterm)
        ]
        disconnected_signal_pins = [
            iterm.getMTerm().getName()
            for iterm in iterms
            if not is_special(iterm) and not is_connected(iterm)
        ]
        table.add_row(
            instance.getName(),
            ",".join(power_pins),
            ",".join(disconnected_power_pins),
            ",".join(signal_pins),
            ",".join(disconnected_signal_pins),
        )
        disconnected_pins_count += len(disconnected_signal_pins) + len(
            disconnected_power_pins
        )
    if disconnected_pins_count > 0:
        rich.print(table)

    print(f"Found {disconnected_pins_count} disconnected pin(s).")

    utl.metric_integer("design__disconnected_pins__count", disconnected_pins_count)


if __name__ == "__main__":
    main()



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import os
import subprocess
from typing import List, Tuple, Optional, Type, Dict, Union

from .flow import Flow, FlowException, FlowError
from ..state import State
from ..steps import (
    Step,
    StepError,
    StepException,
    DeferredStepError,
)
from ..logging import info, success, err


class SequentialFlow(Flow):
    """
    The simplest Flow, running each Step as a stage, serially,
    with nothing happening in parallel and no significant inter-step
    processing.

    All subclasses of this flow have to do is override the :attr:`.Steps` abstract property
    and it would automatically handle the rest. See `Classic` in Built-in Flows for an example.

    :param Substitute: Substitute all instances of one `Step` type by another `Step`
        type in the :attr:`.Steps` attribute for this instance only.

        You may also use the string Step IDs in place of a `Step` type object.

    :param args: Arguments for :class:`Flow`.
    :param kwargs: Keyword arguments for :class:`Flow`.
    """

    @classmethod
    def make(Self, step_ids: List[str]) -> Type[SequentialFlow]:
        Step_list = []
        for name in step_ids:
            step = Step.get(name)
            if step is None:
                raise TypeError(f"No step found with id '{name}'")
            Step_list.append(step)

        class CustomSequentialFlow(SequentialFlow):
            name = "Custom Sequential Flow"
            Steps = Step_list

        return CustomSequentialFlow

    def __init__(
        self,
        Substitute: Optional[
            Dict[str, Union[str, Type[Step]]],
        ] = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        if substitute := Substitute:
            for key, item in substitute.items():
                self._substitute_step(key, item)

    def _substitute_step(
        self,
        id: str,
        with_step: Union[str, Type[Step]],
    ):
        step_index: Optional[int] = None
        for i, step in enumerate(self.Steps):
            if step.id.lower() == id.lower():
                step_index = i
                break
        if step_index is None:
            raise FlowException(
                f"Failed to process start step '{step}': no step with ID '{step}' found in flow."
            )

        if isinstance(with_step, str):
            with_step_opt = Step.get(with_step)
            if with_step_opt is None:
                raise FlowException(
                    f"Could not substitute '{step.id}' with '{with_step}': no step with ID '{with_step}' found."
                )
            with_step = with_step_opt

        self.Steps[i] = with_step

    def run(
        self,
        initial_state: State,
        frm: Optional[str] = None,
        to: Optional[str] = None,
        skip: Optional[List[str]] = None,
        **kwargs,
    ) -> Tuple[State, List[Step]]:
        step_ids = {cls.id.lower(): cls.id for cls in reversed(self.Steps)}
        skipped_ids = []

        frm_resolved = None
        if frm is not None:
            if id := step_ids.get(frm.lower()):
                frm_resolved = id
            else:
                raise FlowException(
                    f"Failed to process start step '{frm}': no step with ID '{frm}' found in flow."
                )

        to_resolved = None
        if to is not None:
            if id := step_ids.get(to.lower()):
                to_resolved = id
            else:
                raise FlowException(
                    f"Failed to process start step '{to}': no step with ID '{to}' found in flow."
                )

        if skipped_steps := skip:
            for skipped_step in skipped_steps:
                if id := step_ids.get(skipped_step.lower()):
                    skipped_ids.append(id)
                else:
                    raise FlowException(
                        f"Failed to process skipped step '{skipped_step}': no step with ID '{skipped_step}' found in flow."
                    )
        step_count = len(self.Steps)
        self.set_max_stage_count(step_count)

        step_list = []

        info("Starting…")

        executing = frm is None
        deferred_errors = []

        current_state = initial_state
        for cls in self.Steps:
            step = cls(config=self.config, state_in=current_state, flow=self)
            if frm_resolved is not None and frm_resolved == step.id:
                executing = True

            self.start_stage(step.name)
            increment_ordinal = True
            if not executing or cls.id in skipped_ids:
                info(f"Skipping step '{step.name}'…")
                increment_ordinal = False
            else:
                step_list.append(step)
                try:
                    current_state = step.start()
                except StepException as e:
                    raise FlowException(str(e))
                except DeferredStepError as e:
                    deferred_errors.append(str(e))
                except (StepError, subprocess.CalledProcessError) as e:
                    raise FlowError(str(e))

            self.end_stage(increment_ordinal=increment_ordinal)

            if to_resolved and to_resolved == step.id:
                executing = False
        if len(deferred_errors) != 0:
            err("The following deferred step errors have been encountered:")
            for error in deferred_errors:
                err(error)
            raise FlowError("One or more deferred errors were encountered.")

        assert self.run_dir is not None
        final_views_path = os.path.join(self.run_dir, "final")
        info(f"Saving final views to '{final_views_path}'…")
        try:
            current_state.save_snapshot(final_views_path)
        except Exception as e:
            raise FlowException(f"Failed to save final views: {e}")
        success("Flow complete.")
        return (current_state, step_list)



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import os
import json
import shutil
from decimal import Decimal
from collections import UserString
from typing import Iterator, List, Mapping, TypeVar, Union, Optional, Dict, Any

from .design_format import DesignFormat, DesignFormatObject
from ..common import mkdirp


class Path(UserString, os.PathLike):
    """
    A Path type for OpenLane configuration variables.

    Basically just a string.
    """

    def __fspath__(self) -> str:
        return str(self)

    def exists(self) -> bool:
        """
        A convenience method calling :meth:`os.path.exists`
        """
        return os.path.exists(self)


class InvalidState(RuntimeError):
    pass


class StateEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, Decimal):
            if o.as_integer_ratio()[1] == 1:
                return int(o)
            else:
                return float(o)
        elif isinstance(o, Path):
            return str(o)
        return super(StateEncoder, self).default(o)


VT = Union[Path, Dict[str, Path], None]


class State(Mapping[str, VT]):
    """
    Basically, a dictionary with keys of type :class:`DesignFormat` and values
    of (nested dictionaries of) :class:`Path`.

    The state is the only thing that can be altered by steps other than the
    filesystem.

    :attr metrics: A dictionary that carries statistics about the design: area,
        wire length, et cetera, but also miscellaneous data, for example, whether
        it passed a certain check or not.
    """

    _data: dict

    metrics: dict

    def __init__(self, metrics: Optional[dict] = None) -> None:
        super().__init__()
        self._data = {}
        for format in DesignFormat:
            self[format] = None
        self.metrics = metrics or {}

    def __getitem__(self, key: Union[DesignFormat, str]) -> VT:
        if isinstance(key, DesignFormat):
            id: str = key.value.id
            key = id
        return self._data[key]

    def __setitem__(self, key: Union[DesignFormat, str], item: VT):
        if isinstance(key, DesignFormat):
            id: str = key.value.id
            key = id
        self._data[key] = item

    def _as_dict(self, metrics: bool = True) -> Dict[str, Any]:
        final: Dict[Any, Any] = dict(self)
        if metrics:
            final["metrics"] = self.metrics
        return final

    T = TypeVar("T", Dict, List)

    def __copy_recursive__(self, input: T) -> T:
        def resolve_value(value):
            value_final = value
            if isinstance(value, dict):
                value_final = self.__copy_recursive__(value)
            elif isinstance(value, list):
                value_final = self.__copy_recursive__(value)
            return value_final

        if isinstance(input, list):
            result = []
            for value in input:
                result.append(resolve_value(value))
            return result
        else:
            result = {}
            for key, value in input.items():
                result[key] = resolve_value(value)
            return result

    def copy(self: "State") -> "State":
        new = State()
        new._data = self.__copy_recursive__(self._data)
        new.metrics = self.__copy_recursive__(self.metrics)
        return new

    def __iter__(self) -> Iterator[str]:
        return iter(self._data)

    def keys(self):
        return self._data.keys()

    def values(self):
        return self._data.values()

    def items(self):
        return self._data.items()

    def __len__(self) -> int:
        return len(self._data)

    def __repr__(self) -> str:
        return self._as_dict().__repr__()

    def dumps(self, **kwargs) -> str:
        """
        Dumps data as JSON.
        """
        if "indent" not in kwargs:
            kwargs["indent"] = 4
        return json.dumps(self._as_dict(), cls=StateEncoder, **kwargs)

    def _save_snapshot_recursive(
        self,
        path: Union[str, os.PathLike],
        views: Union[Dict, "State"],
        key_path: str = "",
    ):
        mkdirp(path)
        for key, value in views.items():
            current_key_path = f"{key_path}.{key}"
            if value is None:
                continue
            current_folder = key.strip("*")
            if df := DesignFormat.by_id(key):
                # For type-checker: all guaranteed to be DesignFormatObjects
                assert isinstance(df.value, DesignFormatObject)
                current_folder = df.value.folder

            if isinstance(value, dict):
                subdirectory = os.path.join(path, current_folder)
                self._save_snapshot_recursive(
                    subdirectory,
                    value,
                    key_path=current_key_path,
                )
            else:
                target_dir = os.path.join(
                    path,
                    current_folder,
                )
                mkdirp(target_dir)
                target_path = os.path.join(target_dir, os.path.basename(value))
                shutil.copyfile(value, target_path, follow_symlinks=True)

    def save_snapshot(self, path: Union[str, os.PathLike]):
        """
        Validates the current state then saves all views to a folder by
        design format, including the metrics.

        :param path: The folder that would contain other folders.
        """
        self.validate()
        self._save_snapshot_recursive(path, self)
        metrics_path = os.path.join(path, "metrics.csv")
        with open(metrics_path, "w") as f:
            f.write("Metric,Value\n")
            for metric in self.metrics:
                f.write(f"{metric}, {self.metrics[metric]}\n")

    def _validate_recursive(
        self,
        views: Dict,
        key_path: str = "",
        depth: int = 0,
    ):
        for key, value in views.items():
            current_key_path = f"{key_path}.{key}"
            if depth == 0 and DesignFormat.by_id(key) is None:
                raise InvalidState(
                    f"Key {current_key_path} does not match a known design format."
                )
            if value is not None:
                if isinstance(value, Path):
                    if not value.exists():
                        raise InvalidState(
                            f"Path for format {current_key_path} does not exist: '{value}'."
                        )
                elif isinstance(value, dict):
                    self._validate_recursive(
                        value, key_path=current_key_path, depth=depth + 1
                    )
                else:
                    raise InvalidState(
                        f"Value for format '{current_key_path}' is not a Path nor a dictionary of strings to Paths: '{value}'."
                    )

    def validate(self):
        """
        Ensures that all paths exist in a State.
        """
        self._validate_recursive(self._as_dict(metrics=False))

    @classmethod
    def _loads_recursive(
        Self,
        target: Union[Dict, "State"],
        views: Union[Dict, "State"],
        validate_path: bool = True,
        key_path: str = "",
    ):
        for key, value in views.items():
            current_key_path = f"{key_path}.{key}"
            if value is None:
                target[key] = value
                continue

            if isinstance(value, dict):
                all_values: Dict = {}
                Self._loads_recursive(
                    all_values,
                    value,
                    validate_path,
                    key_path=current_key_path,
                )
                target[key] = all_values
            else:
                if validate_path and not os.path.exists(value):
                    raise ValueError(
                        f"Provided path '{value}' to design format '{current_key_path}' does not exist."
                    )
                target[key] = Path(value)

    @classmethod
    def loads(Self, json_in: str, validate_path: bool = True) -> "State":
        raw = json.loads(json_in, parse_float=Decimal)

        metrics = raw.get("metrics")
        if metrics is not None:
            del raw["metrics"]

        state = Self(metrics=metrics)

        Self._loads_recursive(state, raw, validate_path)

        return state

    def _repr_html_(self) -> str:
        result = """
        <div style="display: grid; grid-auto-columns: minmax(0, 1fr); grid-auto-rows: minmax(0, 1fr); grid-auto-flow: column;">
            <table style="grid-column-start: 1; grid-column-end: 2;">
                <tr>
                    <th>Format</th>
                    <th>Path</th>
                </tr>
        """
        for id, value in self._as_dict(metrics=False).items():
            if value is None:
                continue

            format = DesignFormat.by_id(id)
            assert format is not None

            value_rel = os.path.relpath(value, ".")

            result += f"""
                <tr>
                    <td>{format.value.id}</td>
                    <td><a href="{value_rel}">{value_rel}</a></td>
                </tr>
            """

        result += """
            </table>
        </div>
        """

        return result



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
import os
import shlex
import inspect
from enum import Enum
from decimal import Decimal, InvalidOperation
from typing import Iterable, Literal, get_origin, get_args
from dataclasses import _MISSING_TYPE, MISSING, dataclass, field, fields, is_dataclass
from typing import Union, Type, List, Optional, Tuple, Any, Dict, Callable, Sequence

from .config import Config
from .resolve import process_string, Keys as SpecialKeys
from ..state import Path

# Scalar = Union[Type[str], Type[Decimal], Type[Path], Type[bool]]
# VType = Union[Scalar, List[Scalar]]


class zip_first(object):
    """
    Works like ``zip_longest`` but always stops when the first iterant stops.
    """

    def __init__(self, a: Iterable, b: Iterable, fillvalue: Any) -> None:
        self.a = a
        self.b = b
        self.fillvalue = fillvalue

    def __iter__(self):
        self.iter_a = iter(self.a)
        self.iter_b = iter(self.b)
        return self

    def __next__(self):
        a = next(self.iter_a)
        b = self.fillvalue
        try:
            b = next(self.iter_b)
        except StopIteration:
            pass
        return (a, b)


newline_rx = re.compile(r"\n")


def is_optional(t: Type[Any]) -> bool:
    type_args = get_args(t)
    return get_origin(t) is Union and type(None) in type_args


def some_of(t: Type[Any]) -> Type[Any]:
    if not is_optional(t):
        return t

    # t must be a Union with None if we're here

    type_args = get_args(t)
    if len(type_args) == 2:
        # Either (Type | None) or (None | Type)
        if type_args[0] != Type[None]:
            return type_args[0]
        else:
            return type_args[1]
    else:
        return t


def repr_type(t: Type[Any]) -> str:
    optional = is_optional(t)
    some = some_of(t)

    type_string = some.__name__
    if inspect.isclass(some) and issubclass(some, Enum):
        type_string = "｜".join([str(e.name) for e in some])
        type_string = f"`{type_string}`"
    else:
        origin, args = get_origin(some), get_args(some)
        if origin is not None:
            if origin == Union:
                arg_strings = [repr_type(arg) for arg in args]
                type_string = "｜".join(arg_strings)
                type_string = f"({type_string})"
            elif origin == Literal:
                return str(args[0])
            else:
                arg_strings = [repr_type(arg) for arg in args]
                type_string = f"{type_string}[{','.join(arg_strings)}]"

    return type_string + ("?" if optional else "")


@dataclass
class Variable:
    """
    An object representing a configuration variable for a PDK, a Flow or a Step.

    :param name: A string name for the Variable. Because of backwards compatility
        with OpenLane 1, the convention is ``UPPER_SNAKE_CASE``.

    :param type: A Python type object representing the variable.

        Supported scalars:

        - ``int``
        - ``decimal.Decimal``
        - ``bool``
        - ``str``
        - :class:`Path`

        Supported products:

        - ``Union`` (incl. ``Optional``)
        - ``List``
        - ``Tuple``
        - ``Dict``
        - ``Enum``

        Other:

        - ``dataclass`` types composed of the above.

    :param description: A human-readable description of the variable. Used to
        generate help strings and documentation.

    :param default: A default value for the variable.

        Optional variables have an implicit default value of ``None``.

    :param deprecated_names: A list of deprecated names for said variable.

        An element of the list can alternative be a tuple of a name and a Callable
        used to perform a translation for when a renamed variable is also slightly
        modified.

    :param units: Used only in documentation: the unit corresponding to this
        object, i.e., μm, pF, etc. Can be any string, but for consistency, SI units
        must be represented in terms of their official symbols.
    """

    name: str
    type: Any
    description: str
    default: Any = None
    deprecated_names: List[Union[str, Tuple[str, Callable]]] = field(
        default_factory=list
    )

    units: Optional[str] = None

    @property
    def optional(self) -> bool:
        """
        Returns whether a variable's type is an `Option type <https://en.wikipedia.org/wiki/Option_type>`_.
        """
        return is_optional(self.type)

    @property
    def some(self) -> Any:
        """
        Returns the type of a variable presuming it is not None.

        If a variable is not Optional, that is simply the type specified in the
        :ivar:`type` field.
        """
        return some_of(self.type)

    def type_repr_md(self) -> str:
        """
        Prints a pretty Markdown string representation of the Variable's type.
        """
        return repr_type(self.type)

    def desc_repr_md(self) -> str:
        """
        Prints the description, but with newlines escaped for Markdown.
        """
        return newline_rx.sub("<br />", self.description)

    def _process(
        self,
        key_path: str,
        value: Any,
        values_so_far: Dict[str, Any],
        validating_type: Type[Any],
        default: Any = None,
        explicitly_specified: bool = True,
    ):
        if explicitly_specified and value is None:
            # User explicitly specified "null" for this value: only error if
            # value is not optional
            if not is_optional(validating_type):
                raise ValueError(
                    f"Non-optional variable {key_path} received a null value."
                )
            else:
                return None
        elif not explicitly_specified and value is None:
            # User did not specify a value for this variable: couple outcomes
            if default is not None:
                return self._process(
                    key_path=key_path,
                    value=default,
                    validating_type=validating_type,
                    values_so_far=values_so_far,
                )
            elif not is_optional(validating_type):
                raise ValueError(
                    f"Required variable {key_path} did not get a specified value."
                )
            else:
                return None

        assert (
            explicitly_specified and value is not None
        ), "Configurator has built an inconsistent state."

        if is_optional(validating_type):
            validating_type = some_of(validating_type)

        if type(value) == str:
            value = process_string(value, values_so_far)

        type_origin = get_origin(validating_type)
        type_args = get_args(validating_type)
        if type_origin in [list, tuple]:
            return_value = list()
            raw = value
            if isinstance(raw, list):
                pass
            elif isinstance(raw, str):
                if "," in raw:
                    raw = raw.split(",")
                elif ";" in raw:
                    raw = raw.split(";")
                else:
                    raw = raw.split()
            else:
                raise ValueError(
                    f"Invalid List provided for variable {key_path}: {value}"
                )

            if type_origin == tuple:
                if len(raw) != len(type_args):
                    raise ValueError(
                        f"Invalid {validating_type} provided for variable {key_path}: ({len(raw)}/{len(type_args)}) tuple entries provided"
                    )

            for i, (item, value_type) in enumerate(
                zip_first(raw, type_args, fillvalue=type_args[0])
            ):
                return_value.append(
                    self._process(
                        key_path=f"{key_path}[{i}]",
                        value=item,
                        validating_type=value_type,
                        values_so_far=values_so_far,
                    )
                )

            if type_origin == tuple:
                return tuple(raw)

            return return_value
        elif type_origin == dict:
            raw = value
            key_type, value_type = type_args
            if isinstance(raw, dict):
                pass
            elif isinstance(raw, str):
                # Assuming Tcl format:
                components = shlex.split(value)
                if len(components) % 2 != 0:
                    raise ValueError(
                        f"Tcl-style flat dictionary provided for variable {key_path} is invalid: uneven number of components ({len(components)})"
                    )
                raw = {}
                for i in range(0, len(components) // 2):
                    key = components[2 * i]
                    val = components[2 * i + 1]
                    raw[key] = val
            else:
                raise ValueError(
                    f"Value provided for variable {key_path} of type {validating_type} is not a dictionary: '{value}'"
                )

            processed = {}
            for key, val in raw.items():
                key_validated = self._process(
                    key_path=key_path,
                    value=key,
                    validating_type=key_type,
                    values_so_far=values_so_far,
                )
                value_validated = self._process(
                    key_path=f"{key_path}.{key_validated}",
                    value=val,
                    validating_type=value_type,
                    values_so_far=values_so_far,
                )
                processed[key_validated] = value_validated

            return processed
        elif type_origin == Union:
            final_value = None
            errors = []
            for arg in type_args:
                try:
                    final_value = self._process(
                        key_path=key_path,
                        value=value,
                        validating_type=arg,
                        values_so_far=values_so_far,
                    )
                except ValueError as e:
                    errors.append(f"\t{str(e)}")
            if final_value is not None:
                return final_value
            else:
                raise ValueError(
                    "\n".join(
                        [
                            f"Value for {key_path} is invalid for union {repr_type(validating_type)}:"
                        ]
                        + errors
                    )
                )
        elif type_origin == Literal:
            arg = type_args[0]
            if value == arg:
                return value
            else:
                raise ValueError(f"Value for {key_path} is not '{arg}': '{value}'")
        elif is_dataclass(validating_type):
            raw = value
            if not isinstance(raw, dict):
                raise ValueError(
                    f"Value provided for deserializable path {validating_type} at {key_path} is not a dictionary."
                )
            kwargs_dict = {}
            for current_field in fields(validating_type):
                key = current_field.name
                subtype = current_field.type
                explicitly_specified = False
                if key in raw:
                    explicitly_specified = True
                field_value = raw.get(key)
                field_default = None
                if (
                    current_field.default is not None
                    and type(current_field.default) != _MISSING_TYPE
                ):
                    field_default = current_field.default
                if current_field.default_factory != MISSING:
                    field_default = current_field.default_factory()
                value_processed = self._process(
                    key_path=f"{key_path}.{key}",
                    value=field_value,
                    explicitly_specified=explicitly_specified,
                    default=field_default,
                    validating_type=subtype,
                    values_so_far=values_so_far,
                )
                kwargs_dict[key] = value_processed
            return validating_type(**kwargs_dict)
        elif validating_type == Path:
            if not os.path.exists(str(value)):
                raise ValueError(
                    f"Path provided for variable {key_path} does not exist: '{value}'"
                )
            return Path(value)
        elif validating_type == bool:
            if value in ["1", "true", True]:
                return True
            elif value in ["0", "false", False]:
                return False
            else:
                raise ValueError(
                    f"Value provided for variable {key_path} of type {validating_type} is invalid: '{value}'"
                )
        elif issubclass(validating_type, Enum):
            try:
                return validating_type[value]
            except KeyError:
                raise ValueError(
                    f"Variable provided for variable {key_path} of enumerated type {validating_type} is invalid: '{value}'"
                )
        elif issubclass(validating_type, Decimal):
            try:
                return Decimal(value)
            except InvalidOperation:
                raise ValueError(
                    f"Value provided for variable {key_path} of type {validating_type} is invalid: '{value}'"
                )

        else:
            try:
                return validating_type(value)
            except ValueError as e:
                raise ValueError(
                    f"Value provided for variable {key_path} of type {validating_type} is invalid: '{value}' {e}"
                )

    def _compile(
        self,
        mutable_config: Config,
        warning_list_ref: List[str],
        values_so_far: Dict[str, Any],
    ) -> Any:
        exists, value = mutable_config.extract(self.name)

        i = 0
        while (
            not exists
            and self.deprecated_names is not None
            and i < len(self.deprecated_names)
        ):
            deprecated_name = self.deprecated_names[i]
            deprecated_callable = lambda x: x
            if not isinstance(deprecated_name, str):
                deprecated_name, deprecated_callable = deprecated_name
            exists, value = mutable_config.extract(deprecated_name)
            if exists:
                warning_list_ref.append(
                    f"The configuration variable '{deprecated_name}' is deprecated. Please check the docs for the usage on the replacement variable '{self.name}'."
                )
            if value is not None:
                value = deprecated_callable(value)
            i = i + 1

        return self._process(
            key_path=self.name,
            value=value,
            default=self.default,
            validating_type=self.type,
            values_so_far=values_so_far,
            explicitly_specified=exists,
        )

    def __hash__(self) -> int:
        return hash((self.name, self.type, self.default))

    def __eq__(self, rhs: object) -> bool:
        if not isinstance(rhs, Variable):
            raise NotImplementedError()
        return (
            self.name == rhs.name
            and self.type == rhs.type
            and self.default == rhs.default
        )

    @classmethod
    def process_config(
        Self,
        config: Config,
        variables: Sequence["Variable"],
        removed: Optional[Dict[str, str]] = None,
    ) -> Tuple[Config, List[str], List[str]]:
        """
        Verifies a configuration object against a list of variables, returning
        an object with the variables normalized according to their types.

        :param config: The input, raw configuration object.
        :param variables: A sequence or some other iterable of variables.
        :param removed: A dictionary of variables that may have existed at a point in
            time, but then have gotten removed. Useful to give feedback to the user.
        :returns: A tuple of:
            [0] A final, processed configuration.
            [1] A list of warnings.
            [2] A list of errors.

            If the third element is non-empty, the first object is invalid.
        """
        if removed is None:
            removed = {}

        warnings: List[str] = []
        errors = []
        final = Config()._unlock()
        mutable = config.copy()._unlock()

        # Special Deprecation Behaviors
        if (
            mutable.get("DIODE_INSERTION_STRATEGY") is not None
        ):  # Can't use := because 0 is a valid value
            _, dis = mutable.extract("DIODE_INSERTION_STRATEGY")
            try:
                dis = int(dis)
            except ValueError:
                pass
            if not isinstance(dis, int) or dis in [1, 2, 5] or dis > 6:
                errors.append(
                    f"DIODE_INSERTION_STRATEGY '{dis}' is not available in OpenLane 2. See 'Migrating DIODE_INSERTION_STRATEGY' in the docs for more info."
                )
            else:
                warnings.append(
                    "The DIODE_INSERTION_STRATEGY variable has been deprecated. See 'Migrating DIODE_INSERTION_STRATEGY' in the docs for more info."
                )

                final["GRT_REPAIR_ANTENNAS"] = False
                final["RUN_HEURISTIC_DIODE_INSERTION"] = False
                final["DIODE_ON_PORTS"] = "none"
                if dis in [3, 6]:
                    final["GRT_REPAIR_ANTENNAS"] = True
                if dis in [5, 6]:
                    final["RUN_HEURISTIC_DIODE_INSERTION"] = True
                    final["DIODE_ON_PORTS"] = "in"

        # Macros
        translated_macros = False
        if mutable.get("EXTRA_SPEFS") is not None:
            mutable["MACROS"] = mutable.get("MACROS") or {}

            _, extra_spef_list = mutable.extract("EXTRA_SPEFS")
            if isinstance(extra_spef_list, str):
                extra_spef_list = extra_spef_list.split(" ")

            if not isinstance(extra_spef_list, list):
                errors.append(
                    f"Invalid type for 'EXTRA_SPEFS': {type(extra_spef_list)}. It is recommended that you update your configuration to use the Macro object."
                )
            elif len(extra_spef_list) % 4 != 0:
                errors.append(
                    "Invalid value for 'EXTRA_SPEFS': Element count not divisible by four. It is recommended that you update your configuration to use the Macro object."
                )
            else:
                translated_macros = True
                warnings.append(
                    "The configuration variable 'EXTRA_SPEFS' is deprecated. Check the docs on how to use the new 'MACROS' configuration variable."
                )
                for i in range(len(extra_spef_list) // 4):
                    start = i * 4
                    module, min, nom, max = (
                        extra_spef_list[start],
                        extra_spef_list[start + 1],
                        extra_spef_list[start + 2],
                        extra_spef_list[start + 3],
                    )
                    macro_dict = {"module": module, "gds": ["/dev/null"]}
                    macro_dict["spef"] = {
                        "min_*": [min],
                        "nom_*": [nom],
                        "max_*": [max],
                    }
                    mutable["MACROS"][module] = macro_dict

        for variable in variables:
            try:
                value_processed = variable._compile(
                    mutable_config=mutable,
                    warning_list_ref=warnings,
                    values_so_far=final.data,
                )
                final[variable.name] = value_processed
            except ValueError as e:
                errors.append(str(e))

        for key in sorted(mutable.keys()):
            if key in vars(SpecialKeys).values():
                continue
            if key in removed:
                warnings.append(f"'{key}' has been removed: {removed[key]}")
            elif "_OPT" not in key:
                warnings.append(f"Unknown key '{key}' provided.")

        if translated_macros:
            for macro in final["MACROS"].values():
                if macro.gds == "/dev/null":
                    macro.gds = Path("")

        return (final._lock(), warnings, errors)



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import re
from decimal import Decimal
from typing import List, Optional, Dict

from .variable import Variable
from .config import Config
from ..state import Path

# Note that values in this file do not take defaults.

pdk_variables = [
    Variable(
        "PDK_ROOT",
        str,
        "The path to all PDKs. This doesn't actually need to be defined by the PDK config file itself.",
    ),
    # Core/Common
    Variable(
        "STD_CELL_LIBRARY",
        str,
        "Specifies the default standard cell library to be used under the specified PDK.",
    ),
    Variable(
        "VDD_PIN",
        str,
        "The power pin for the cells.",
    ),
    Variable(
        "GND_PIN",
        str,
        "The ground pin for the cells.",
    ),
    Variable(
        "WIRE_LENGTH_THRESHOLD",
        Optional[Decimal],
        "A value above which wire lengths generate warnings.",
        units="µm",
    ),
    Variable(
        "TECH_LEFS",
        Dict[str, Path],
        "Map of corner patterns to to technology LEF files. A corner not matched here will not be supported by OpenRCX in the default flow.",
    ),
    Variable(
        "CELL_LEFS",
        List[Path],
        "Path(s) to the cell LEF file(s).",
        deprecated_names=["CELLS_LEF"],
    ),
    Variable(
        "CELL_GDS",
        List[Path],
        "Path(s) to the cell GDSII file(s).",
        deprecated_names=["GDS_FILES", "CELLS_GDS"],
    ),
    Variable(
        "GPIO_PADS_LEF",
        Optional[List[Path]],
        "Path(s) to GPIO pad LEF file(s).",
    ),
    Variable(
        "GPIO_PADS_LEF_CORE_SIDE",
        Optional[List[Path]],
        "Path(s) to GPIO pad LEF file(s) as used for routing (?).",
    ),
    Variable(
        "GPIO_PADS_VERILOG",
        Optional[List[Path]],
        "Path(s) to GPIO pad Verilog models.",
    ),
    Variable(
        "GPIO_PADS_PREFIX",
        Optional[List[str]],
        "A list of pad cell name prefixes.",
    ),
    # Signoff
    ## Netgen
    Variable(
        "NETGEN_SETUP",
        Optional[Path],
        "A path to the setup file for Netgen used to configure LVS. If set to None, this PDK will not support Netgen-based steps.",
        deprecated_names=["NETGEN_SETUP_FILE"],
    ),
    ## Magic
    Variable(
        "MAGICRC",
        Optional[Path],
        "A path to the `.magicrc` file which is sourced before running magic in the flow.",
        deprecated_names=["MAGIC_MAGICRC"],
    ),
    Variable(
        "MAGIC_TECH",
        Optional[Path],
        "A path to a Magic tech file which, mainly, has DRC rules.",
        deprecated_names=["MAGIC_TECH_FILE"],
    ),
    ## Klayout,
    Variable(
        "KLAYOUT_TECH",
        Optional[Path],
        "A path to the KLayout layer technology (.lyt) file.",
    ),
    Variable(
        "KLAYOUT_PROPERTIES",
        Optional[Path],
        "A path to the KLayout layer properties (.lyp) file.",
    ),
    Variable(
        "KLAYOUT_DRC_TECH_SCRIPT",
        Optional[Path],
        "A path to a KLayout DRC tech script.",
    ),
    Variable(
        "KLAYOUT_DEF_LAYER_MAP",
        Optional[Path],
        "A path to the KLayout LEF/DEF layer mapping (.map) file.",
    ),
    Variable(
        "KLAYOUT_XOR_IGNORE_LAYERS",
        Optional[List[str]],
        "KLayout layers to ignore during XOR operations.",
    ),
    ## Timing and Power
    Variable(
        "DEFAULT_MAX_TRAN",
        Optional[Decimal],
        "Defines the default maximum transition value used in Synthesis and CTS.\nA minimum of 0.1 * CLOCK_PERIOD and this variable, if defined, is used.",
        units="ns",
    ),
    Variable(
        "WIRE_RC_LAYER",
        Optional[str],
        "A metal layer with which to estimate parasitics in earlier stages of the flow.",
    ),
    Variable(
        "RCX_RULESETS",
        Optional[Dict[str, Path]],
        "Map of corner patterns to OpenRCX extraction rules. A corner not matched by exactly one pattern in this dictionary will not be supported by OpenRCX, and a PDK not specifying this variable will not be supported by OpenRCX.",
    ),
    Variable(
        "DEFAULT_CORNER",
        str,
        "The interconnect/process/voltage/temperature corner (IPVT) to use the characterized lib files compatible with by default.",
    ),
    Variable(
        "STA_CORNERS",
        List[str],
        "A list of fully qualified IPVT (Interconnect, transistor Process, Voltage, and Temperature) timing corners on which to conduct multi-corner static timing analysis.",
    ),
    # Floorplanning
    Variable(
        "FP_TRACKS_INFO",
        Path,
        "A path to the a classic OpenROAD `.tracks` file. Used by the floorplanner to generate tracks.",
        deprecated_names=["TRACKS_INFO_FILE"],
    ),
    Variable(
        "FP_TAPCELL_DIST",
        Decimal,
        "The distance between tap cell columns.",
        units="µm",
    ),
    Variable(
        "FP_PDN_RAIL_OFFSET",
        Decimal,
        "The offset for the power distribution network rails for first metal layer.",
        units="µm",
    ),
    Variable(
        "FP_PDN_VWIDTH",
        Decimal,
        "The strap width for the vertical layer in generated power distribution networks.",
        units="µm",
    ),
    Variable(
        "FP_PDN_VSPACING",
        Decimal,
        "The spacing between vertical straps in generated power distribution networks.",
        units="µm",
    ),
    Variable(
        "FP_PDN_HSPACING",
        Decimal,
        "The spacing between horizontal straps in generated power distribution networks.",
        units="µm",
    ),
    Variable(
        "FP_PDN_HWIDTH",
        Decimal,
        "The strap width for the horizontal layer in generated power distribution networks.",
        units="µm",
    ),
    Variable(
        "FP_PDN_CORE_RING_VWIDTH",
        Decimal,
        "The width for the vertical layer in the core ring of generated power distribution networks.",
        units="µm",
    ),
    Variable(
        "FP_PDN_CORE_RING_HWIDTH",
        Decimal,
        "The width for the horizontal layer in the core ring of generated power distribution networks.",
        units="µm",
    ),
    Variable(
        "FP_PDN_CORE_RING_VSPACING",
        Decimal,
        "The spacing for the vertical layer in the core ring of generated power distribution networks.",
        units="µm",
    ),
    Variable(
        "FP_PDN_CORE_RING_HSPACING",
        Decimal,
        "The spacing for the horizontal layer in the core ring of generated power distribution networks.",
        units="µm",
    ),
    Variable(
        "FP_PDN_CORE_RING_VOFFSET",
        Decimal,
        "The offset for the vertical layer in the core ring of generated power distribution networks.",
        units="µm",
    ),
    Variable(
        "FP_PDN_CORE_RING_HOFFSET",
        Decimal,
        "The offset for the horizontal layer in the core ring of generated power distribution networks.",
        units="µm",
    ),
    Variable(
        "FP_IO_HLAYER",
        str,
        "The metal layer on which to place horizontal IO pins, i.e., the top and bottom of the die.",
    ),
    Variable(
        "FP_IO_VLAYER",
        str,
        "The metal layer on which to place vertial IO pins, i.e., the top and bottom of the die.",
    ),
    Variable("RT_MIN_LAYER", str, "The lowest metal layer to route on."),
    Variable("RT_MAX_LAYER", str, "The highest metal layer to route on."),
]

scl_variables = [
    # Common
    Variable(
        "SCL_GROUND_PINS",
        List[str],
        "SCL-specific ground pins",
        deprecated_names=["STD_CELL_GROUND_PINS"],
    ),
    Variable(
        "SCL_POWER_PINS",
        List[str],
        "SCL-specific power pins",
        deprecated_names=["STD_CELL_POWER_PINS"],
    ),
    Variable(
        "FILL_CELL",
        List[str],
        "A list of cell names or wildcards of fill cells to be used in fill insertion.",
    ),
    Variable(
        "DECAP_CELL",
        List[str],
        "A list of cell names or wildcards of decap cells to be used in fill insertion.",
    ),
    Variable(
        "LIB",
        Dict[str, List[Path]],
        "A map from corner patterns to a list of associated liberty files. Exactly one entry must match the `DEFAULT_CORNER`.",
    ),
    Variable(
        "SYNTH_EXCLUSION_CELL_LIST",
        Path,
        "Path to a text file containing a list of cells to be excluded from the lib file in synthesis alone. If not defined, the original lib file will be used as-is.",
        deprecated_names=["NO_SYNTH_CELL_LIST"],
    ),
    Variable(
        "PNR_EXCLUSION_CELL_LIST",
        Path,
        "Path to a text file containing a list of undesirable or bad (DRC-failed or complex pinout) cells to be excluded from synthesis AND PnR. If not defined, all cells will be used.",
        deprecated_names=["DRC_EXCLUDE_CELL_LIST"],
    ),
    # Synthesis
    Variable(
        "SYNTH_CAP_LOAD",
        Decimal,
        "Defines the capacitive load on the output ports.",
        units="fF",
    ),
    Variable(
        "SYNTH_DRIVING_CELL",
        str,
        "The cell to drive the input ports, used in synthesis and static timing analysis, in the format `{cell}/{port}`.",
    ),
    Variable(
        "SYNTH_CLK_DRIVING_CELL",
        Optional[str],
        "The cell to drive the clock input ports, used in synthesis and static timing analysis, in the format `{cell}/{port}`. If not specified, `SYNTH_DRIVING_CELL` will be used.",
    ),
    Variable(
        "SYNTH_TIEHI_CELL",
        str,
        "Defines the tie high cell followed by the port that implements the tie high functionality, in the format `{cell}/{port}`.",
    ),
    Variable(
        "SYNTH_TIELO_CELL",
        str,
        "Defines the tie high cell followed by the port that implements the tie low functionality, in the format `{cell}/{port}`.",
    ),
    Variable(
        "SYNTH_BUFFER_CELL",
        str,
        "Defines a buffer port to be used by yosys during synthesis: in the format `{cell}/{input_port}/{output_port}`",
    ),
    ## Mappings
    Variable(
        "SYNTH_LATCH_MAP",
        Optional[Path],
        "A path to a file contianing the latch mapping for Yosys.",
    ),
    Variable(
        "SYNTH_TRISTATE_MAP",
        Optional[Path],
        "A path to a file contianing the tri-state buffer mapping for Yosys.",
        deprecated_names=["TRISTATE_BUFFER_MAP"],
    ),
    Variable(
        "SYNTH_CSA_MAP",
        Optional[Path],
        "A path to a file containing the carry-select adder mapping for Yosys.",
        deprecated_names=["CARRY_SELECT_ADDER_MAP"],
    ),
    Variable(
        "SYNTH_RCA_MAP",
        Optional[Path],
        "A path to a file containing the ripple-carry adder mapping for Yosys.",
        deprecated_names=["RIPPLE_CARRY_ADDER_MAP"],
    ),
    Variable(
        "SYNTH_FA_MAP",
        Optional[Path],
        "A path to a file containing the full adder mapping for Yosys.",
        deprecated_names=["FULL_ADDER_MAP"],
    ),
    Variable(
        "SYNTH_MUX_MAP",
        Optional[Path],
        "A path to a file containing the mux mapping for Yosys.",
    ),
    Variable(
        "SYNTH_MUX4_MAP",
        Optional[Path],
        "A path to a file containing the mux4 mapping for Yosys.",
    ),
    # Clock Tree Synthesis
    Variable(
        "CTS_ROOT_BUFFER",
        str,
        "Defines the cell inserted at the root of the clock tree. Used in CTS.",
    ),
    Variable(
        "CTS_CLK_BUFFERS",
        List[str],
        "Defines the list of clock buffers to be used in CTS.",
        deprecated_names=["CTS_CLK_BUFFER_LIST"],
    ),
    Variable(
        "CTS_MAX_CAP",
        Decimal,
        "Defines the maximum capacitance, used in CTS.",
        units="pF",
    ),
    # Floorplanning
    Variable(
        "FP_WELLTAP_CELL",
        str,
        "Defines the cell used for tap insertion.",
    ),
    Variable(
        "FP_ENDCAP_CELL",
        str,
        "Defines so-called 'end-cap' cells- decap cells placed at either sides of a design.",
    ),
    Variable(
        "FP_PDN_RAILS_LAYER",
        str,
        "Defines the metal layer used for PDN rails.",
    ),
    Variable(
        "FP_PDN_RAIL_WIDTH",
        Decimal,
        "Defines the width of PDN rails on the `FP_PDN_RAILS_LAYER` layer.",
        units="µm",
    ),
    Variable(
        "FP_PDN_UPPER_LAYER",
        str,
        "Defines the upper PDN layer.",
    ),
    Variable(
        "FP_PDN_LOWER_LAYER",
        str,
        "Defines the lower PDN layer.",
    ),
    Variable(
        "IGNORE_DISCONNECTED_MODULES",
        Optional[List[str]],
        "Modules (or cells) to ignore when checking for disconnected pins.",
    ),
    # Placement
    Variable(
        "PLACE_SITE",
        str,
        "Defines the main placement site in placement as specified in the technology LEF files, to generate the placement grid.",
    ),
    Variable(
        "PLACE_SITE_WIDTH",
        Decimal,
        "The site width for the previously designated place site.",
        units="µm",
    ),
    Variable(
        "PLACE_SITE_HEIGHT",
        Decimal,
        "The site height for the previously designated place site.",
        units="µm",
    ),
    Variable(
        "GPL_CELL_PADDING",
        Decimal,
        "Cell padding value (in sites) for global placement. The number will be integer divided by 2 and placed on both sides.",
        units="sites",
    ),
    Variable(
        "DPL_CELL_PADDING",
        Decimal,
        "Cell padding value (in sites) for detailed placement. The number will be integer divided by 2 and placed on both sides. Should be <= global placement.",
        units="sites",
    ),
    Variable(
        "CELL_PAD_EXCLUDE",
        List[str],
        "Defines a list of cells to be excluded from cell padding.",
    ),
    # Antenna
    Variable(
        "DIODE_CELL",
        Optional[str],
        "Defines a diode cell used to fix antenna violations, in the format {name}/{port}.",
    ),
    # Routing
    Variable(
        "GRT_LAYER_ADJUSTMENTS",
        List[Decimal],
        "Layer-specific reductions in the routing capacity of the edges between the cells in the global routing graph, delimited by commas. Values range from 0 through 1.",
    ),
    # CVC
    Variable(
        "CVC_SCRIPTS_DIR",
        Optional[Path],
        "Path to a directory of Circuit Validity Checker (CVC) scripts for the relevant PDK. Must contain the following set of files: `cvcrc`, an initialization file, `cdl.awk`, an awk script to remove black box definitions from SPICE files, `models`, cell models, and finally `power.awk`, an awk script that adds power information to the verilog netlists.\nIf this path is not defined, this PDK will be marked incompatible with CVC.",
    ),
]


def migrate_old_config(config: Config) -> Config:
    new = config.copy()._unlock()
    # 1. Migrate SYNTH_DRIVING_CELL
    del new["SYNTH_DRIVING_CELL"]
    del new["SYNTH_DRIVING_CELL_PIN"]
    new[
        "SYNTH_DRIVING_CELL"
    ] = f"{config['SYNTH_DRIVING_CELL']}/{config['SYNTH_DRIVING_CELL_PIN']}"

    # 2. Migrate SYNTH_TIE{HI,LO}_CELL
    del new["SYNTH_TIEHI_PORT"]
    new["SYNTH_TIEHI_CELL"] = "/".join(config["SYNTH_TIEHI_PORT"].split(" "))

    del new["SYNTH_TIELO_PORT"]
    new["SYNTH_TIELO_CELL"] = "/".join(config["SYNTH_TIELO_PORT"].split(" "))

    # 3. Migrate SYNTH_BUFFER_CELL
    del new["SYNTH_MIN_BUF_PORT"]
    new["SYNTH_BUFFER_CELL"] = "/".join(config["SYNTH_MIN_BUF_PORT"].split(" "))

    # 4. Migrate DIODE_CELL
    del new["DIODE_CELL"]
    del new["DIODE_CELL_PIN"]
    new["DIODE_CELL"] = f"{config['DIODE_CELL']}/{config['DIODE_CELL_PIN']}"

    # 5. Interconnect Corners
    del new["RCX_RULES"]
    new["RCX_RULESETS"] = f"nom_* \"{config['RCX_RULES']}\""
    if config.get("RCX_RULES_MIN") is not None:
        del new["RCX_RULES_MIN"]
        new["RCX_RULESETS"] += f" min_* \"{config['RCX_RULES_MIN']}\""
    if config.get("RCX_RULES_MAX") is not None:
        del new["RCX_RULES_MAX"]
        new["RCX_RULESETS"] += f" max_* \"{config['RCX_RULES_MAX']}\""

    del new["TECH_LEF"]
    new["TECH_LEFS"] = f"nom_* \"{config['TECH_LEF']}\""
    if config.get("TECH_LEF_MIN") is not None:
        del new["TECH_LEF_MIN"]
        new["TECH_LEFS"] += f" min_* \"{config['TECH_LEF_MIN']}\""
    if config.get("TECH_LEF_MAX") is not None:
        del new["TECH_LEF_MAX"]
        new["TECH_LEFS"] += f" max_* \"{config['TECH_LEF_MAX']}\""

    # 6. Timing Corners
    lib_sta: Dict[str, List[str]] = {}
    ws = re.compile(r"\s+")

    default_pvt = ""

    def process_sta(key: str):
        nonlocal new, default_pvt
        lib_raw = new.pop(key)
        if lib_raw is None:
            return
        lib = lib_raw.strip()
        lib_list = ws.split(lib)
        first_lib = os.path.basename(lib_list[0])[:-4]
        pvt = first_lib.split("__")[1]
        if default_pvt == "":
            default_pvt = pvt
        corner = f"*_{pvt}"
        lib_sta[corner] = lib_list

    process_sta("LIB_SYNTH")
    process_sta("LIB_SLOWEST")
    process_sta("LIB_FASTEST")

    if new["PDK"].startswith("sky130"):
        new["STA_CORNERS"] = [
            "nom_tt_025C_1v80",
            "nom_ss_100C_1v60",
            "nom_ff_n40C_1v95",
            "min_tt_025C_1v80",
            "min_ss_100C_1v60",
            "min_ff_n40C_1v95",
            "max_tt_025C_1v80",
            "max_ss_100C_1v60",
            "max_ff_n40C_1v95",
        ]
    elif new["PDK"].startswith("gf180mcu"):
        new["STA_CORNERS"] = [
            "nom_tt_025C_5v00",
            "nom_ss_125C_4v50",
            "nom_ff_n40C_5v50",
            "min_tt_025C_5v00",
            "min_ss_125C_4v50",
            "min_ff_n40C_5v50",
            "max_tt_025C_5v00",
            "max_ss_125C_4v50",
            "max_ff_n40C_5v50",
        ]

    new["DEFAULT_CORNER"] = f"nom_{default_pvt}"
    new["LIB"] = lib_sta

    # 7. Disconnected Modules (sky130)
    if new["PDK"].startswith("sky130"):
        new["IGNORE_DISCONNECTED_MODULES"] = "sky130_fd_sc_hd__conb_1"

    # 8. Invalid Variables (gf180mcu)
    if new["PDK"].startswith("gf180mcu"):
        del new["GPIO_PADS_LEF"]
        del new["GPIO_PADS_VERILOG"]

        del new["CARRY_SELECT_ADDER_MAP"]
        del new["FULL_ADDER_MAP"]
        del new["RIPPLE_CARRY_ADDER_MAP"]
        del new["SYNTH_LATCH_MAP"]
        del new["TRISTATE_BUFFER_MAP"]

        del new["KLAYOUT_DRC_TECH_SCRIPT"]

        new[
            "SYNTH_CLK_DRIVING_CELL"
        ] = f"{config['SYNTH_CLK_DRIVING_CELL']}/{config['SYNTH_DRIVING_CELL_PIN']}"
    return new._lock()


all_variables: List[Variable] = pdk_variables + scl_variables
removed_variables: Dict[str, str] = {
    "FAKEDIODE_CELL": "Fake diode-based strategies have been removed.",
    "LIB_SYNTH": "Use the lib dictionary instead.",
}



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import subprocess
import sys
from base64 import b64encode
from typing import Optional, List

from .step import Step, StepError, StepException
from ..logging import warn
from ..state import DesignFormat, State, Path
from ..config import Variable, Config
from ..common import get_script_dir
from ..utils import Toolbox


def get_lef_args(config: Config, toolbox: Toolbox) -> List[str]:
    tech_lefs = toolbox.filter_views(config, config["TECH_LEFS"])
    if len(tech_lefs) != 1:
        raise StepException(
            "Misconfigured SCL: 'TECH_LEFS' must return exactly one Tech LEF for its default timing corner."
        )

    lef_args = [
        "--input-lef",
        str(tech_lefs[0]),
    ]

    for lef in config["CELL_LEFS"]:
        lef_args.append("--input-lef")
        lef_args.append(str(lef))

    macro_lefs = toolbox.get_macro_views(config, DesignFormat.LEF)
    for lef in macro_lefs:
        lef_args.append("--input-lef")
        lef_args.append(str(lef))

    if extra_lefs := config["EXTRA_LEFS"]:
        for lef in extra_lefs:
            lef_args.append("--input-lef")
            lef_args.append(str(lef))

    return lef_args


@Step.factory.register()
class StreamOut(Step):
    """
    Converts DEF views into GDSII streams using KLayout.

    The PDK must support KLayout for this step to work, otherwise
    it will be skipped.

    If ``PRIMARY_SIGNOFF_TOOL`` is set to ``"klayout"``, both GDS and KLAYOUT_GDS
    will be updated, and if set to another tool, only ``KLAYOUT_GDS`` will be
    updated.
    """

    id = "KLayout.StreamOut"
    name = "GDSII Stream Out (KLayout)"
    flow_control_variable = "RUN_KLAYOUT_STREAMOUT"

    inputs = [DesignFormat.DEF]
    outputs = [DesignFormat.GDS, DesignFormat.KLAYOUT_GDS]

    config_vars = [
        Variable(
            "RUN_KLAYOUT_STREAMOUT",
            bool,
            "Enables streaming GDSII using KLayout.",
            default=True,
            deprecated_names=["RUN_KLAYOUT"],
        )
    ]

    def run(self, state_in: State, **kwargs) -> State:
        state_out = super().run(state_in, **kwargs)

        lyp = self.config["KLAYOUT_PROPERTIES"]
        lyt = self.config["KLAYOUT_TECH"]
        lym = self.config["KLAYOUT_DEF_LAYER_MAP"]
        if None in [lyp, lyt, lym]:
            if self.config["PRIMARY_SIGNOFF_TOOL"].value == "klayout":
                raise StepError(
                    "One of KLAYOUT_PROPERTIES, KLAYOUT_TECH or KLAYOUT_DEF_LAYER_MAP is unset, yet, KLayout is set as the primary sign-off tool."
                )
            warn(
                "One of KLAYOUT_PROPERTIES, KLAYOUT_TECH or KLAYOUT_DEF_LAYER_MAP is unset. Returning state unaltered…"
            )
            return state_out

        klayout_gds_out = os.path.join(
            self.step_dir,
            f"{self.config['DESIGN_NAME']}.{DesignFormat.KLAYOUT_GDS.value.extension}",
        )

        layout_args = []
        layout_args += get_lef_args(self.config, self.toolbox)

        for gds in self.config["CELL_GDS"]:
            layout_args.append("--with-gds-file")
            layout_args.append(gds)
        for gds in self.toolbox.get_macro_views(self.config, DesignFormat.GDS):
            layout_args.append("--with-gds-file")
            layout_args.append(gds)
        if extra_gds := self.config["EXTRA_GDS_FILES"]:
            for gds in extra_gds:
                layout_args.append("--with-gds-file")
                layout_args.append(gds)

        kwargs, env = self.extract_env(kwargs)

        self.run_subprocess(
            [
                sys.executable,
                os.path.join(
                    get_script_dir(),
                    "klayout",
                    "stream_out.py",
                ),
                state_in[DesignFormat.DEF.value.id],
                "--output",
                klayout_gds_out,
                "--lyt",
                lyt,
                "--lyp",
                lyp,
                "--lym",
                lym,
                "--top",
                self.config["DESIGN_NAME"],
            ]
            + layout_args,
            env=env,
        )

        state_out[DesignFormat.KLAYOUT_GDS] = Path(klayout_gds_out)

        if self.config["PRIMARY_SIGNOFF_TOOL"].value == "klayout":
            state_out[DesignFormat.GDS] = state_out[DesignFormat.KLAYOUT_GDS]

        return state_out

    def layout_preview(self) -> Optional[str]:
        if self.state_out is None:
            return None
        assert self.toolbox is not None

        if image := self.toolbox.render_png(
            self.config, str(self.state_out["klayout_gds"])
        ):
            image_encoded = b64encode(image).decode("utf8")
            return f'<img src="data:image/png;base64,{image_encoded}" />'

        return None


@Step.factory.register()
class XOR(Step):
    """
    Performs an XOR operation on the Magic and KLayout GDS views. The idea is:
    if there's any difference between the GDSII streams between the two tools,
    one of them have it wrong and that may lead to ambiguity.
    """

    id = "KLayout.XOR"
    name = "KLayout vs. Magic XOR"
    flow_control_variable = "RUN_KLAYOUT_XOR"

    inputs = [
        DesignFormat.MAG_GDS,
        DesignFormat.KLAYOUT_GDS,
    ]

    config_vars = [
        Variable(
            "RUN_KLAYOUT_XOR",
            bool,
            "Enables running KLayout XOR on the two GDSII files generated by Magic and Klayout. Stream-outs for both KLayout and Magic should have already run, and the PDK must support both signoff tools.",
            default=True,
        ),
        Variable(
            "KLAYOUT_XOR_THREADS",
            int,
            "Specifies number of threads used in the KLayout XOR check.",
            default=1,
        ),
    ]

    def run(self, state_in: State, **kwargs) -> State:
        state_out = super().run(state_in, **kwargs)

        ignored = ""
        if ignore_list := self.config["KLAYOUT_XOR_IGNORE_LAYERS"]:
            ignored = ";".join(ignore_list)

        layout_a = state_out[DesignFormat.MAG_GDS]
        if layout_a is None:
            warn("No Magic stream-out has been performed. Skipping XOR process…")
            return state_out
        layout_b = state_out[DesignFormat.KLAYOUT_GDS]
        if layout_b is None:
            warn("No KLayout stream-out has been performed. Skipping XOR process…")
            return state_out

        kwargs, env = self.extract_env(kwargs)

        self.run_subprocess(
            [
                "ruby",
                os.path.join(
                    get_script_dir(),
                    "klayout",
                    "xor.drc",
                ),
                "--output",
                os.path.join(self.step_dir, "xor.xml"),
                "--top",
                self.config["DESIGN_NAME"],
                "--ignore",
                ignored,
                layout_a,
                layout_b,
            ],
            env=env,
        )

        state_out.metrics["design__xor_difference__count"] = int(
            open(os.path.join(self.step_dir, "difference_count.rpt")).read().strip()
        )

        return state_out


@Step.factory.register()
class OpenGUI(Step):
    """
    Opens the DEF view in the KLayout GUI, with layers loaded and mapped
    properly. Useful to inspect ``.klayout.xml`` database files and the like.
    """

    id = "KLayout.OpenGUI"
    name = "Open In GUI"

    inputs = [DesignFormat.DEF]
    outputs = []

    def run(self, state_in: State, **kwargs) -> State:
        state_out = super().run(state_in, **kwargs)

        lyp = self.config["KLAYOUT_PROPERTIES"]
        lyt = self.config["KLAYOUT_TECH"]
        lym = self.config["KLAYOUT_DEF_LAYER_MAP"]
        if None in [lyp, lyt, lym]:
            raise StepError(
                "Cannot open design in KLayout as the PDK does not appear to support KLayout."
            )

        lefs = get_lef_args(self.config, self.toolbox)
        kwargs, env = self.extract_env(kwargs)

        cmd = [
            sys.executable,
            os.path.join(
                get_script_dir(),
                "klayout",
                "open_design.py",
            ),
            "--lyt",
            str(lyt),
            "--lyp",
            str(lyp),
            "--lym",
            str(lym),
            str(state_out.get("def")),
        ] + lefs

        print(" ".join(cmd))

        subprocess.check_call(
            cmd,
            env=env,
        )

        return state_out



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# flake8: noqa
"""
The OpenLane API

These modules serve as the infrastructure for OpenLane-based flows.
Using these functions, users may either utilize built-in flows, or build custom
Flows and/or custom Steps for the design of more sophisticated chips.

..
    no-imported-members
"""

from .config import Variable, Config, ConfigBuilder, InvalidConfig
from .flows import Flow, SequentialFlow
from .state import State, DesignFormat
from .steps import Step
from .common import *
from .plugins import discovered_plugins
from .__version__ import __version__



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from decimal import Decimal
from typing import Optional, List

from ..config import Variable

io_layer_variables = [
    Variable(
        "FP_IO_VEXTEND",
        Decimal,
        "Extends the vertical io pins outside of the die by the specified units.",
        default=0,
        units="μm",
    ),
    Variable(
        "FP_IO_HEXTEND",
        Decimal,
        "Extends the horizontal io pins outside of the die by the specified units.",
        default=0,
        units="μm",
    ),
    Variable(
        "FP_IO_VLENGTH",
        Decimal,
        "The length of the vertical IOs.",
        default=4,
        units="µm",
    ),
    Variable(
        "FP_IO_HLENGTH",
        Decimal,
        "The length of the horizontal IOs.",
        default=4,
        units="µm",
    ),
    Variable(
        "FP_IO_VTHICKNESS_MULT",
        Decimal,
        "A multiplier for vertical pin thickness. Base thickness is the pins layer minwidth.",
        default=2,
    ),
    Variable(
        "FP_IO_HTHICKNESS_MULT",
        Decimal,
        "A multiplier for horizontal pin thickness. Base thickness is the pins layer minwidth.",
        default=2,
    ),
]

pdn_variables = [
    Variable(
        "FP_PDN_VOFFSET",
        Decimal,
        "The offset of the vertical power stripes on the metal layer 4 in the power distribution network.",
        default=16.32,
        units="μm",
    ),
    Variable(
        "FP_PDN_VPITCH",
        Decimal,
        "The pitch of the vertical power stripes on the metal layer 4 in the power distribution network.",
        default=153.6,
        units="μm",
    ),
    Variable(
        "FP_PDN_HOFFSET",
        Decimal,
        "The offset of the horizontal power stripes on the metal layer 5 in the power distribution network.",
        default=16.65,
        units="μm",
    ),
    Variable(
        "FP_PDN_HPITCH",
        Decimal,
        "The pitch of the horizontal power stripes on the metal layer 5 in the power distribution network.",
        default=153.18,
        units="μm",
    ),
    Variable(
        "FP_PDN_AUTO_ADJUST",
        bool,
        "Decides whether or not the flow should attempt to re-adjust the power grid, in order for it to fit inside the core area of the design, if needed.",
        default=True,
    ),
    Variable(
        "FP_PDN_SKIPTRIM",
        bool,
        "Enables `-skip_trim` option during pdngen which skips the metal trim step, which attempts to remove metal stubs.",
        default=True,
    ),
    Variable(
        "FP_PDN_CORE_RING",
        bool,
        "Enables adding a core ring around the design. More details on the control variables in the PDK config documentation.",
        default=False,
    ),
    Variable(
        "FP_PDN_ENABLE_RAILS",
        bool,
        "Enables the creation of rails in the power grid.",
        default=True,
    ),
    Variable(
        "FP_PDN_CHECK_NODES",
        bool,
        "Enables checking for unconnected nodes in the power grid.",
        default=True,
    ),
    Variable(
        "FP_PDN_HORIZONTAL_HALO",
        str,
        "Sets the horizontal halo around the macros during power grid insertion.",
        default=10,
        units="µm",
    ),
    Variable(
        "FP_PDN_VERTICAL_HALO",
        Decimal,
        "Sets the vertical halo around the macros during power grid insertion.",
        default="expr::$FP_PDN_HORIZONTAL_HALO",
        units="µm",
    ),
    Variable(
        "DESIGN_IS_CORE",
        bool,
        "Controls the layers used in the power grid. Depending on whether the design is the core of a chip or a macro inside the core.",
        default=True,
    ),
]

routing_layer_variables = [
    Variable(
        "RT_CLOCK_MIN_LAYER",
        Optional[str],
        "The name of lowest layer to be used in routing the clock net.",
    ),
    Variable(
        "RT_CLOCK_MAX_LAYER",
        Optional[str],
        "The name of highest layer to be used in routing the clock net.",
    ),
    Variable(
        "GRT_ADJUSTMENT",
        Decimal,
        "Reduction in the routing capacity of the edges between the cells in the global routing graph for all layers. Values range from 0 to 1.  1 = most reduction, 0 = least reduction.",
        default=0.3,
    ),
    Variable(
        "GRT_MACRO_EXTENSION",
        int,
        "Sets the number of GCells added to the blockages boundaries from macros. A GCell is typically defined in terms of Mx routing tracks. The default GCell size is 15 M3 pitches.",
        default=0,
    ),
]


dpl_variables = [
    Variable(
        "PL_OPTIMIZE_MIRRORING",
        bool,
        "Specifies whether or not to run an optimize_mirroring pass whenever detailed placement happens. This pass will mirror the cells whenever possible to optimize the design.",
        default=True,
    ),
    Variable(
        "PL_MAX_DISPLACEMENT_X",
        Decimal,
        "Specifies how far an instance can be moved along the X-axis when finding a site where it can be placed during detailed placement.",
        default=500,
        units="μm",
    ),
    Variable(
        "PL_MAX_DISPLACEMENT_Y",
        Decimal,
        "Specifies how far an instance can be moved along the Y-axis when finding a site where it can be placed during detailed placement.",
        default=100,
        units="μm",
    ),
]


rsz_variables = dpl_variables + [
    Variable(
        "RSZ_DONT_TOUCH_RX",
        str,
        'A single regular expression designating nets as "don\'t touch" by resizer optimizations.',
        default="$^",
        deprecated_names=["UNBUFFER_NETS"],
    ),
    Variable(
        "RSZ_DONT_USE_CELLS",
        Optional[List[str]],
        "An optional list of cells to not use during resizer optimizations.",
        deprecated_names=["DONT_USE_CELLS"],
    ),
]

constraint_variables = [
    Variable(
        "MAX_FANOUT_CONSTRAINT",
        int,
        "The max load that the output ports can drive to be used as a constraint on Synthesis and CTS.",
        default=10,
        units="cells",
        deprecated_names=["SYNTH_MAX_FANOUT"],
    ),
    Variable(
        "MAX_TRANSITION_CONSTRAINT",
        Optional[Decimal],
        "The max transition time (slew) from high to low or low to high on cell inputs in ns to be used as a constraint on Synthesis and CTS. If not provided, it is calculated at runtime as `10%` of the provided clock period, unless that exceeds the PDK's `DEFAULT_MAX_TRAN` value.",
        units="ns",
        deprecated_names=["SYNTH_MAX_TRAN"],
    ),
    Variable(
        "CLOCK_UNCERTAINTY_CONSTRAINT",
        Decimal,
        "Specifies a value for the clock uncertainty/jitter for timing analysis.",
        default=0.25,
        units="ns",
        deprecated_names=["SYNTH_CLOCK_UNCERTAINTY"],
    ),
    Variable(
        "CLOCK_TRANSITION_CONSTRAINT",
        Decimal,
        "Specifies a value for the clock transition/slew for timing analysis.",
        default=0.15,
        units="ns",
        deprecated_names=["SYNTH_CLOCK_TRANSITION"],
    ),
    Variable(
        "TIME_DERATING_CONSTRAINT",
        Decimal,
        "Specifies a derating factor to multiply the path delays with. It specifies the upper and lower ranges of timing.",
        default=5,
        units="%",
        deprecated_names=["SYNTH_TIMING_DERATE"],
    ),
    Variable(
        "IO_DELAY_CONSTRAINT",
        Decimal,
        "Specifies the percentage of the clock period used in the input/output delays.",
        default=20,
        units="%",
        deprecated_names=["IO_PCT"],
    ),
]



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import json
import yaml
from enum import Enum
from decimal import Decimal
from textwrap import dedent
from dataclasses import dataclass, asdict, is_dataclass
from collections import UserDict
from typing import (
    Any,
    ClassVar,
    Tuple,
    Union,
    List,
    Optional,
    Sequence,
)

from ..state import Path


def StringEnum(name: str, values: Sequence[str]):
    """
    Creates a string enumeration where the keys and values are the same.
    """
    return Enum(name, [(value, value) for value in values])


@dataclass
class Meta:
    version: int = 1
    flow: Union[None, str, List[str]] = "Classic"


class ConfigEncoder(json.JSONEncoder):
    def default(self, o):
        if is_dataclass(o):
            return asdict(o)
        elif isinstance(o, Decimal):
            if o.as_integer_ratio()[1] == 1:
                return int(o)
            else:
                return float(o)
        elif isinstance(o, Path):
            return str(o)
        elif isinstance(o, Enum):
            return o.value
        return super(ConfigEncoder, self).default(o)


class Config(UserDict):
    """
    A map from OpenLane configuration variable keys to their values.

    It is recommended that you construct these using the :class:`ConfigBuilder`
    singleton class.

    A created Config is immutable and cannot be modified by Steps.
    """

    current_interactive: ClassVar[Optional["Config"]] = None
    locked: bool = False

    meta: Meta = Meta(version=1)
    interactive: bool = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.locked = True

    def _unlock(self) -> "Config":
        self.locked = False
        return self

    def _lock(self) -> "Config":
        self.locked = True
        return self

    @classmethod
    def get_meta(
        Self,
        json_config_in: Union[str, os.PathLike],
        flow_override: Optional[str] = None,
    ) -> Optional[Meta]:
        """
        Returns the Meta object of a JSON configuration file

        :param config_in: A configuration file.
        :returns: Either a Meta object, or if the file is not a JSON file, None.
        """
        try:
            obj = json.load(open(json_config_in, encoding="utf8"))
        except (json.JSONDecodeError, IsADirectoryError):
            return None

        meta = Meta()
        if meta_raw := obj.get("meta"):
            meta = Meta(**meta_raw)

        if flow_override is not None:
            meta.flow = flow_override

        return meta

    def copy(self, **overrides) -> "Config":
        """
        Produces a shallow copy of the configuration object.

        :param overrides: A series of configuration overrides as key-value pairs.
            These values are NOT validated and you should not be overriding these
            haphazardly.
        """
        import copy

        data = self.data
        try:
            self.data = {}
            c = copy.copy(self)
        finally:
            self.data = data
        c._unlock()
        c.update(self)
        c.update(overrides)
        c._lock()
        return c

    def __setitem__(self, key: str, item: Any):
        if self.locked:
            raise AttributeError("Config objects are immutable.")

        return super().__setitem__(key, item)

    def dumps(self) -> str:
        """
        :returns: A JSON string representing the configuration object.
        """
        data = self.data.copy()
        data["meta"] = self.meta
        return json.dumps(data, indent=2, cls=ConfigEncoder, sort_keys=True)

    def check(self, key: str) -> Tuple[bool, Any]:
        """
        :param key: A string key representing an OpenLane configuration variable.
        :returns: A tuple of whether that key exists, and if so, its value.

            Do note ``None`` is a valid value, so simply checking if the
            second element ``is not None`` is invalid to check whether
            a key exists.
        """
        return (key in self.keys(), self.get(key))

    def extract(self, key: str) -> Tuple[bool, Any]:
        """
        A mutating method that attempts to find a key, and, if it exists,
        deletes it from the configuration object and returns its value.

        :param key: A string key representing an OpenLane configuration variable.
        :returns: A tuple of whether that key existed, and if so its value.

            Do note ``None`` is a valid value, so simply checking if the
            second element ``is not None`` is invalid to check whether
            a key exists.
        """
        found, value = self.check(key)
        if found:
            del self[key]
        return (found, value)

    def _repr_markdown_(self) -> str:
        title = "Interactive Configuration" if self.interactive else "Configuration"
        values_title = "Initial Values" if self.interactive else "Values"
        return (
            dedent(
                f"""
                ### {title}
                #### {values_title}

                <br />

                ```yml
                %s
                ```
                """
            )
            % yaml.safe_dump(json.loads(self.dumps()))
        )



# Copyright 2023 Efabless Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The Step Module

This modules includes various functions for importing and/or generating OpenLane
configuration objects. Configuration objects are the primary input to a flow.
"""
from .step import (
    Step,
    DeferredStepError,
    StepError,
    StepException,
)
from .tclstep import TclStep
from . import checker as Checker

# You'll notice some TclStep subclasses are exposed separately-
# this is for documentation.
from . import yosys as Yosys
from .yosys import YosysStep

from . import openroad as OpenROAD
from .openroad import OpenROADStep

from . import magic as Magic
from .magic import MagicStep

from . import odb as Odb

# from .odb import OdbpyStep

from . import netgen as Netgen

# from .netgen import NetgenStep

from . import klayout as KLayout
from . import misc as Misc
